import datetime
from typing import TYPE_CHECKING, ClassVar

from django.db import models
from django.utils import timezone
from django.utils.translation import gettext_lazy as _

from allianceauth.eveonline.evelinks import eveimageserver

from corptools.utils import to_roman_numeral

from ..managers import (
    EveCategoryManager, EveGroupManager, EveItemTypeManager, EveMoonManager,
    EveNameManager, EvePlanetManager,
)
from .utils import JSONModel

if TYPE_CHECKING:
    from esi.stubs import (
        UniverseCategoriesCategoryIdGet, UniverseGroupsGroupIdGet,
        UniverseTypesTypeIdGet, UniverseTypesTypeIdGet_Dogma_attributesItem,
    )


class EveName(models.Model):
    objects: ClassVar[EveNameManager] = EveNameManager()

    eve_id = models.BigIntegerField(primary_key=True)
    name = models.CharField(max_length=255)
    category = models.CharField(max_length=50)

    # optionals for character/corp
    corporation = models.ForeignKey(
        'EveName',
        on_delete=models.SET_NULL,
        null=True,
        default=None,
        related_name="corp"
    )
    alliance = models.ForeignKey(
        'EveName',
        on_delete=models.SET_NULL,
        null=True,
        default=None,
        related_name="alli"
    )
    last_update = models.DateTimeField(auto_now=True)

    CHARACTER = "character"
    CORPORATION = "corporation"
    ALLIANCE = "alliance"

    def __str__(self):
        return self.name

    def get_image_url(self):
        if self.category == self.CHARACTER:
            return eveimageserver.character_portrait_url(self.eve_id)
        elif self.category == self.CORPORATION:
            return eveimageserver.corporation_logo_url(self.eve_id)
        elif self.category == self.ALLIANCE:
            return eveimageserver.alliance_logo_url(self.eve_id)
        elif self.category == 'faction':  # CCP...
            return eveimageserver.corporation_logo_url(self.eve_id)

    def needs_update(self):
        return self.last_update + datetime.timedelta(days=30) < timezone.now()


class EveItemCategory(JSONModel):
    """
    categories.jsonl
    """
    _filename = "categories.jsonl"
    _update_fields = [
        "name",
    ]

    objects: ClassVar[EveCategoryManager] = EveCategoryManager()
    category_id = models.BigIntegerField(primary_key=True)
    name = models.CharField(max_length=255)  # unknown max

    def __str__(self):
        return self.name

    @classmethod
    def from_esi_model(cls, esi_model: "UniverseCategoriesCategoryIdGet"):
        return cls(
            category_id=esi_model.category_id,
            name=esi_model.name
        )

    @classmethod
    def from_jsonl(cls, json_data, names=False):
        return cls(
            category_id=json_data.get("_key"),
            name=json_data.get("name", {}).get("en"),
        )


class EveItemGroup(JSONModel):
    """
    groups.jsonl
    """
    _filename = "groups.jsonl"
    _update_fields = [
        "name",
        "category"
    ]

    objects: ClassVar[EveGroupManager] = EveGroupManager()
    group_id = models.BigIntegerField(primary_key=True)
    name = models.CharField(max_length=255)  # unknown max
    category = models.ForeignKey(
        EveItemCategory, on_delete=models.SET_NULL, null=True, default=None)

    def __str__(self):
        return self.name

    @classmethod
    def from_esi_model(cls, esi_model: "UniverseGroupsGroupIdGet"):
        return cls(
            group_id=esi_model.group_id,
            category_id=esi_model.category_id,
            name=esi_model.name
        )

    @classmethod
    def from_jsonl(cls, json_data, names=False):
        return cls(
            group_id=json_data.get("_key"),
            name=json_data.get("name", {}).get("en"),
            category_id=json_data.get("categoryID")
        )


class EveItemType(JSONModel):
    """
    types.jsonl
    """
    _filename = "types.jsonl"
    _update_fields = [
        "name",
        "description",
        "mass",
        "packaged_volume",
        "portion_size",
        "volume",
        "published",
        "radius",
        "group"
    ]

    objects: ClassVar[EveItemTypeManager] = EveItemTypeManager()
    type_id = models.BigIntegerField(primary_key=True)
    name = models.CharField(max_length=255)  # unknown max
    group = models.ForeignKey(
        EveItemGroup, on_delete=models.SET_NULL, null=True, default=None)
    description = models.TextField(null=True, default=None)
    mass = models.FloatField(null=True, default=None)
    packaged_volume = models.FloatField(null=True, default=None)
    portion_size = models.FloatField(null=True, default=None)
    volume = models.FloatField(null=True, default=None)
    published = models.BooleanField()
    radius = models.FloatField(null=True, default=None)

    def __str__(self):
        return self.name

    @classmethod
    def from_esi_model(cls, esi_model: "UniverseTypesTypeIdGet"):
        return cls(
            type_id=esi_model.type_id,
            group_id=esi_model.group_id,
            name=esi_model.name,
            description=esi_model.description,
            mass=esi_model.mass,
            packaged_volume=esi_model.packaged_volume,
            portion_size=esi_model.portion_size,
            volume=esi_model.volume,
            published=esi_model.published,
            radius=esi_model.radius
        )

    @classmethod
    def from_jsonl(cls, json_data, names=False):
        return cls(
            type_id=json_data.get("_key"),
            name=json_data.get("name", {}).get("en", None),
            description=json_data.get("description", {}).get("en", None),
            group_id=json_data.get("groupID"),
            mass=json_data.get("mass", None),
            packaged_volume=json_data.get("packaged_volume", None),
            portion_size=json_data.get("portion_size", None),
            volume=json_data.get("volume", None),
            published=json_data.get("published", False),
            radius=json_data.get("radius", None)
        )


class EveItemDogmaAttribute(JSONModel):
    """
    typeDogma.jsonl
    """
    _filename = "typeDogma.jsonl"
    _update_fields = False

    eve_type = models.ForeignKey(
        EveItemType, on_delete=models.SET_NULL, null=True, default=None)
    attribute_id = models.BigIntegerField(null=True, default=None)
    value = models.FloatField(null=True, default=None)

    @classmethod
    def from_esi_model(cls, type_id, esi_model: "UniverseTypesTypeIdGet_Dogma_attributesItem"):
        return cls(
            eve_type_id=type_id,
            attribute_id=esi_model.attribute_id,
            value=esi_model.value
        )

    @classmethod
    def from_jsonl(cls, json_data, names=False):
        output = []
        for _attrib in json_data.get("dogmaAttributes", []):
            output.append(
                cls(
                    eve_type_id=json_data.get("_key"),
                    attribute_id=_attrib.get("attributeID"),
                    value=_attrib.get("value")
                )
            )
        return output

    @classmethod
    def load_from_sde(cls, folder_name):
        dogma_query = cls.objects.all()
        if dogma_query.exists():
            # speed and we are not caring about f-keys or signals on these models
            dogma_query._raw_delete(dogma_query.db)
        super().load_from_sde(folder_name)


class InvTypeMaterials(JSONModel):
    """
    typeDogma.jsonl
    """
    _filename = "typeDogma.jsonl"
    _update_fields = False

    eve_type = models.ForeignKey(
        EveItemType,
        on_delete=models.SET_NULL,
        null=True,
        default=None
    )
    qty = models.IntegerField()
    type_id = models.IntegerField()
    material_type_id = models.IntegerField()
    met_type = models.ForeignKey(
        EveItemType,
        on_delete=models.SET_NULL,
        null=True,
        default=None,
        related_name="met_type"
    )

    @classmethod
    def from_jsonl(cls, json_data, names=False):
        output = []
        for _mats in json_data.get("materials", []):
            output.append(
                cls(
                    eve_type_id=json_data.get("_key"),
                    type_id=json_data.get("_key"),
                    material_type_id=_mats.get("materialTypeID"),
                    met_type_id=_mats.get("materialTypeID"),
                    qty=_mats.get("qty")
                )
            )
        return output

    @classmethod
    def load_from_sde(cls, folder_name):
        dogma_query = cls.objects.all()
        if dogma_query.exists():
            # speed and we are not caring about f-keys or signals on these models
            dogma_query._raw_delete(dogma_query.db)
        super().load_from_sde(folder_name)


class MapRegion(JSONModel):
    """
    mapRegions.jsonl
    """
    _filename = "mapRegions.jsonl"
    _update_fields = [
        "name",
        "description"
    ]

    region_id = models.BigIntegerField(primary_key=True)
    name = models.CharField(max_length=255)
    description = models.TextField(null=True, default=None, )

    def __str__(self):
        return self.name

    @classmethod
    def from_jsonl(cls, json_data, names=False):
        return cls(
            region_id=json_data.get("_key"),
            name=json_data.get("name", {}).get("en"),
            description=json_data.get("description", {}).get("en"),
        )


class MapConstellation(JSONModel):
    """
    mapConstellations.jsonl
    """
    _filename = "mapConstellations.jsonl"
    _update_fields = [
        "name",
        "region"
    ]

    constellation_id = models.BigIntegerField(primary_key=True)
    name = models.CharField(max_length=255)
    region = models.ForeignKey(
        MapRegion, on_delete=models.SET_NULL, null=True, default=None)

    def __str__(self):
        return self.name

    @classmethod
    def from_jsonl(cls, json_data, names=False):
        return cls(
            constellation_id=json_data.get("_key"),
            name=json_data.get("name", {}).get("en"),
            region_id=json_data.get("regionID")
        )


class MapSystem(JSONModel):
    """
    mapSolarSystems.jsonl
    """
    _filename = "mapSolarSystems.jsonl"
    _update_fields = [
        "name",
        "security_class",
        "security_status",
        "star_id",
        "constellation",
        "x",
        "y",
        "z",
    ]

    system_id = models.BigIntegerField(primary_key=True)
    security_status = models.FloatField()
    name = models.CharField(max_length=255)
    x = models.FloatField()
    y = models.FloatField()
    z = models.FloatField()
    security_class = models.CharField(max_length=255, null=True, default=None)
    star_id = models.IntegerField(null=True, default=None)
    constellation = models.ForeignKey(
        MapConstellation, on_delete=models.SET_NULL, null=True, default=None)

    def __str__(self):
        return self.name

    @classmethod
    def from_jsonl(cls, json_data, names=False):
        return cls(
            system_id=json_data.get("_key"),
            name=json_data.get("name", {}).get("en"),
            constellation_id=json_data.get("constellationID"),
            star_id=json_data.get("starID"),
            security_class=json_data.get("securityClass"),
            security_status=json_data.get("securityStatus"),
            x=json_data.get("position", {}).get("x"),
            y=json_data.get("position", {}).get("y"),
            z=json_data.get("position", {}).get("z"),
        )


class MapSystemGate(JSONModel):
    """
    mapStargates.jsonl
    """
    _filename = "mapStargates.jsonl"
    _update_fields = False

    from_solar_system = models.ForeignKey(
        MapSystem, on_delete=models.CASCADE, related_name="from_system")
    to_solar_system = models.ForeignKey(
        MapSystem, on_delete=models.CASCADE, related_name="to_system")

    def __str__(self):
        return (self.from_solar_system_id, self.to_solar_system_id)

    @classmethod
    def from_jsonl(cls, json_data, names=False):
        return cls(
            from_solar_system_id=json_data.get("solarSystemID"),
            to_solar_system_id=json_data.get(
                "destination", {}).get("solarSystemID"),
        )

    @classmethod
    def load_from_sde(cls, folder_name):
        gate_qry = cls.objects.all()
        if gate_qry.exists():
            # speed and we are not caring about f-keys or signals on these models
            gate_qry._raw_delete(gate_qry.db)
        super().load_from_sde(folder_name)


class MapSystemPlanet(JSONModel):
    """
    mapPlanets.jsonl
        "system_name planet_roman_numeral"
    """
    _filename = "mapPlanets.jsonl"
    _update_fields = [
        "name",
        "system",
        "x",
        "y",
        "z",
    ]

    objects: ClassVar[EvePlanetManager] = EvePlanetManager()

    planet_id = models.IntegerField(primary_key=True)
    system = models.ForeignKey(
        MapSystem, on_delete=models.CASCADE, related_name="planet")
    name = models.CharField(max_length=255)

    x = models.FloatField()
    y = models.FloatField()
    z = models.FloatField()

    eve_type = models.ForeignKey(
        EveItemType, on_delete=models.SET_NULL, null=True, default=None)

    def __str__(self):
        return (self.name)

    @classmethod
    def name_lookup(cls):
        return {
            s.get("system_id"): s.get("name") for s in
            MapSystem.objects.all().values("system_id", "name")
        }

    @classmethod
    def from_jsonl(cls, json_data, system_names):
        name = f"{system_names[json_data.get('solarSystemID')]} {to_roman_numeral(json_data.get('celestialIndex'))}"
        return cls(
            planet_id=json_data.get("_key"),
            system_id=json_data.get("solarSystemID"),
            name=name,
            x=json_data.get("position", {}).get("x"),
            y=json_data.get("position", {}).get("y"),
            z=json_data.get("position", {}).get("z"),
        )


class MapSystemMoon(JSONModel):
    """
    mapMoons.jsonl
        "system_name planet_roman_numeral - Moon #"
    """
    _filename = "mapMoons.jsonl"
    _update_fields = [
        "name",
        "system",
        "x",
        "y",
        "z",
    ]

    objects: ClassVar[EveMoonManager] = EveMoonManager()

    moon_id = models.IntegerField(primary_key=True)
    system = models.ForeignKey(
        MapSystem, on_delete=models.CASCADE, related_name="moon")
    name = models.CharField(max_length=255)

    x = models.FloatField()
    y = models.FloatField()
    z = models.FloatField()

    def __str__(self):
        return (self.name)

    @classmethod
    def name_lookup(cls):
        return {
            p.get("planet_id"): p.get("name") for p in
            MapSystemPlanet.objects.all().values("planet_id", "name")
        }

    @classmethod
    def from_jsonl(cls, json_data, planet_names):
        name = f"{planet_names[json_data.get('orbitID')]} - Moon {json_data.get('orbitIndex')}"
        return cls(
            moon_id=json_data.get("_key"),
            system_id=json_data.get("solarSystemID"),
            name=name,
            x=json_data.get("position", {}).get("x"),
            y=json_data.get("position", {}).get("y"),
            z=json_data.get("position", {}).get("z"),
        )


class MapJumpBridge(models.Model):
    structure_id = models.BigIntegerField(primary_key=True)
    from_solar_system = models.ForeignKey(
        MapSystem, on_delete=models.CASCADE, related_name="bridge_from_system")
    to_solar_system = models.ForeignKey(
        MapSystem, on_delete=models.CASCADE, related_name="bridge_to_system")
    owner = models.ForeignKey(
        EveName, on_delete=models.SET_NULL, null=True, default=None)
    updated = models.DateTimeField(auto_now=True)
    manually_input = models.BooleanField(default=False)

    def __str__(self):
        return f"{self.from_solar_system.name} >> {self.to_solar_system.name} ({self.structure_id}) (Auto: {not self.manually_input})"


class TypePrice(models.Model):
    item = models.ForeignKey(EveItemType, on_delete=models.DO_NOTHING)
    price = models.DecimalField(max_digits=20, decimal_places=2)
    last_update = models.DateTimeField(auto_now=True)


# ************ is this needed any more?

class OreTax(models.Model):
    item = models.ForeignKey(EveItemType, on_delete=models.DO_NOTHING)
    price = models.DecimalField(max_digits=20, decimal_places=2)
    last_update = models.DateTimeField(auto_now=True)


class OreTaxRates(models.Model):
    tag = models.CharField(max_length=500, default="Mining Tax")
    refine_rate = models.DecimalField(
        max_digits=5, decimal_places=2, default=87.5)
    ore_rate = models.DecimalField(max_digits=5, decimal_places=2)  # normal
    ubiquitous_rate = models.DecimalField(
        max_digits=5, decimal_places=2)  # ubiq
    common_rate = models.DecimalField(max_digits=5, decimal_places=2)  # comon
    uncommon_rate = models.DecimalField(
        max_digits=5, decimal_places=2)  # uncom
    rare_rate = models.DecimalField(max_digits=5, decimal_places=2)  # rare
    excptional_rate = models.DecimalField(
        max_digits=5, decimal_places=2)  # best



import logging
from typing import List

from ninja import NinjaAPI

from allianceauth.services.hooks import get_extension_logger

from corptools import models
from corptools.api import schema

logger = get_extension_logger(__name__)


class FinancesApiEndpoints:

    tags = ["Finances"]

    def __init__(self, api: NinjaAPI):
        @api.get(
            "corporation/wallettypes",
            tags=self.tags
        )
        def get_corporation_wallet_types(request):
            perms = (
                request.user.has_perm('corptools.own_corp_manager')
                | request.user.has_perm('corptools.alliance_corp_manager')
                | request.user.has_perm('corptools.state_corp_manager')
                | request.user.has_perm('corptools.global_corp_manager')
                | request.user.has_perm('corptools.holding_corp_wallets')
            )

            if not perms:
                logging.error(
                    f"Permission Denied for {request.user} to view wallets!")
                return 403, "Permission Denied!"

            ref_types = models.CorporationWalletJournalEntry.objects.values_list(
                "ref_type", flat=True).distinct()

            return 200, ["all"] + list(ref_types)

        @api.get(
            "corporation/{corporation_id}/wallet",
            response={200: List[schema.CorporationWalletEvent], 403: str},
            tags=self.tags
        )
        def get_corporation_wallet(request, corporation_id: int, type_refs: str = "", page: int = 1):
            perms = (
                request.user.has_perm('corptools.own_corp_manager')
                | request.user.has_perm('corptools.alliance_corp_manager')
                | request.user.has_perm('corptools.state_corp_manager')
                | request.user.has_perm('corptools.global_corp_manager')
                | request.user.has_perm('corptools.holding_corp_wallets')
            )

            if not perms:
                logging.error(
                    f"Permission Denied for {request.user} to view wallets!")
                return 403, "Permission Denied!"

            wallet_journal = models.CorporationWalletJournalEntry.get_visible(request.user)\
                .filter(division__corporation__corporation__corporation_id=corporation_id)\
                .select_related('first_party_name', 'second_party_name', 'division')\
                .order_by("-date")

            start_count = (page - 1) * 10000
            end_count = page * 10000

            filter_refs = True
            if type_refs:
                refs = type_refs.split(",")
                if "all" in refs:
                    filter_refs = False

                if len(refs) == 0:
                    return 200, []

                if filter_refs:
                    wallet_journal = wallet_journal.filter(ref_type__in=refs)

            wallet_journal = wallet_journal[start_count:end_count]

            output = []
            for w in wallet_journal:
                output.append(
                    {
                        "division": f"{w.division.division} {w.division.name}",
                        "id": w.entry_id,
                        "date": w.date,
                        "first_party": {
                            "id": w.first_party_id,
                            "name": w.first_party_name.name,
                            "cat": w.first_party_name.category,
                        },
                        "second_party": {
                            "id": w.second_party_id,
                            "name": w.second_party_name.name,
                            "cat": w.second_party_name.category,
                        },
                        "ref_type": w.ref_type,
                        "amount": w.amount,
                        "balance": w.balance,
                        "reason": w.reason,
                    })

            return output

        @api.get(
            "corporation/{corporation_id}/divisions",
            response={200: List, 403: str},
            tags=self.tags
        )
        def get_corporation_divisions(request, corporation_id: int, type_refs: str = "", page: int = 1):
            perms = (
                request.user.has_perm('corptools.own_corp_manager')
                | request.user.has_perm('corptools.alliance_corp_manager')
                | request.user.has_perm('corptools.state_corp_manager')
                | request.user.has_perm('corptools.global_corp_manager')
                | request.user.has_perm('corptools.holding_corp_wallets')
            )

            if not perms:
                logging.error(
                    f"Permission Denied for {request.user} to view wallets!")
                return 403, "Permission Denied!"

            wallet_divisions = models.CorporationWalletDivision.get_visible(
                request.user
            ).filter(
                corporation__corporation__corporation_id=corporation_id
            ).order_by("division")

            output = []
            for w in wallet_divisions:
                output.append(
                    {
                        "division": f"{w.division}",
                        "name": f"{w.name}",
                        "balance": w.balance,
                    })

            return output



from django.core.management.base import BaseCommand

from corptools.models import CharacterWalletJournalEntry
from corptools.tasks import update_wallet_currency


class Command(BaseCommand):
    help = 'Update or Populate EVE Item prices for cat_id 4'

    def handle(self, *args, **options):
        self.stdout.write((
            "Finding all the instances, "
            "This may take a while if you have LOTS of data in your database please wait. "
            "These will then be sent to celery for processing."
        ))
        messages = CharacterWalletJournalEntry.objects.filter(
            reason__contains="@ $"
        )
        total = messages.count()
        self.stdout.write(f"Found {total:,} wallet entries to update")
        if total == 0:
            self.stdout.write(f"Nothing to do... Have a good day!")
            return

        user_input = input("Are you sure you want to proceed? (yes/no) ")
        if user_input == "yes":
            _count = 0
            count = 0
            step = 1000
            for m in messages.values_list("pk", flat=True):
                # update the reason
                update_wallet_currency.apply_async(args=[m], priority=8)
                # lets not spam the shit out of the console
                _count += 1
                if _count >= step:
                    _count = 0
                    count += 1
                    self.stdout.write(f"Sent {count*step:,}/{total:,} Tasks")
            self.stdout.write(f"Completed sending {total:,} tasks!")
        else:
            self.stdout.write("Aborted!")



from django.db import models

from .audits import CharacterAudit, CorporationAudit
from .eve_models import EveItemType


class CharacterIndustryJob(models.Model):
    character = models.ForeignKey(CharacterAudit, on_delete=models.CASCADE)

    activity_id = models.IntegerField()
    blueprint_id = models.BigIntegerField()
    blueprint_location_id = models.BigIntegerField()
    blueprint_type_id = models.BigIntegerField()
    blueprint_type_name = models.ForeignKey(
        EveItemType, on_delete=models.SET_NULL, null=True, default=None, related_name="blueprint_type")
    completed_character_id = models.BigIntegerField(
        default=None, null=True, blank=True)
    completed_date = models.DateTimeField(default=None, null=True, blank=True)
    cost = models.DecimalField(
        max_digits=20, decimal_places=2, default=None, null=True, blank=True)
    duration = models.IntegerField()
    end_date = models.DateTimeField()
    facility_id = models.BigIntegerField()
    installer_id = models.IntegerField()
    job_id = models.IntegerField()
    licensed_runs = models.IntegerField(default=None, null=True, blank=True)
    output_location_id = models.BigIntegerField()
    pause_date = models.DateTimeField(default=None, null=True, blank=True)
    probability = models.FloatField(default=None, null=True, blank=True)
    product_type_id = models.IntegerField()
    product_type_name = models.ForeignKey(
        EveItemType, on_delete=models.SET_NULL, null=True, default=None, related_name="product_type")

    runs = models.IntegerField()
    start_date = models.DateTimeField()
    station_id = models.BigIntegerField()
    status = models.CharField(max_length=15)
    successful_runs = models.IntegerField(default=None, null=True, blank=True)


class CorporationIndustryJob(models.Model):
    """
    https://developers.eveonline.com/api-explorer#/schemas/CorporationsCorporationIdIndustryJobsGet
    """
    corporation = models.ForeignKey(CorporationAudit, on_delete=models.CASCADE)

    activity_id = models.IntegerField()
    blueprint_id = models.BigIntegerField()
    blueprint_location_id = models.BigIntegerField()
    blueprint_type_id = models.BigIntegerField()
    blueprint_type_name = models.ForeignKey(
        EveItemType, on_delete=models.SET_NULL, null=True, default=None, related_name="corp_blueprint_type")
    completed_character_id = models.BigIntegerField(
        default=None, null=True, blank=True)
    completed_date = models.DateTimeField(default=None, null=True, blank=True)
    cost = models.DecimalField(
        max_digits=20, decimal_places=2, default=None, null=True, blank=True)
    duration = models.IntegerField()
    end_date = models.DateTimeField()
    facility_id = models.BigIntegerField()
    installer_id = models.IntegerField()
    job_id = models.IntegerField()
    licensed_runs = models.IntegerField(default=None, null=True, blank=True)
    location_id = models.BigIntegerField()
    output_location_id = models.BigIntegerField()
    pause_date = models.DateTimeField(default=None, null=True, blank=True)
    probability = models.FloatField(default=None, null=True, blank=True)
    product_type_id = models.IntegerField()
    product_type_name = models.ForeignKey(
        EveItemType, on_delete=models.SET_NULL, null=True, default=None, related_name="corp_product_type")

    runs = models.IntegerField()
    start_date = models.DateTimeField()
    status = models.CharField(max_length=15)
    successful_runs = models.IntegerField(default=None, null=True, blank=True)



# Generated by Django 4.2.13 on 2024-05-27 05:28

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('corptools', '0102_corporationaudit_last_change_contracts_and_more'),
    ]

    operations = [
        migrations.CreateModel(
            name='CurrentShipFilter',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=500)),
                ('description', models.CharField(max_length=500)),
                ('count_message_only', models.BooleanField(default=False)),
                ('categories', models.ManyToManyField(blank=True, help_text='Filter on Asset Categories.', to='corptools.eveitemcategory')),
                ('constellations', models.ManyToManyField(blank=True, help_text='Limit filter to specific constellations', to='corptools.mapconstellation')),
                ('groups', models.ManyToManyField(blank=True, help_text='Filter on Asset Groups.', to='corptools.eveitemgroup')),
                ('regions', models.ManyToManyField(blank=True, help_text='Limit filter to specific regions', to='corptools.mapregion')),
                ('systems', models.ManyToManyField(blank=True, help_text='Limit filter to specific systems', to='corptools.mapsystem')),
                ('types', models.ManyToManyField(blank=True, help_text='Filter on Asset Types.', to='corptools.eveitemtype')),
            ],
            options={
                'verbose_name': 'Smart Filter: The Current Ship being Flown',
                'verbose_name_plural': 'Smart Filter: The Current Ship being Flown',
            },
        ),
    ]



from django.core.management.base import BaseCommand

from corptools.tasks import update_category


class Command(BaseCommand):
    help = 'Update or Populate EVE Item Categories'

    def add_arguments(self, parser):
        parser.add_argument('category_ids', nargs='+', type=int)
        parser.add_argument('--inline', action='store_true',
                            help='Run update in this Console not via Celery')

    def handle(self, *args, **options):
        self.stdout.write(f"{options}")

        if len(options["category_ids"]) == 0:
            return self.stdout.write("No Categories!")

        for c in options["category_ids"]:

            if options['inline']:
                self.stdout.write(
                    "Running Tasks inline this may take some time!")
                update_category(c)
            else:
                self.stdout.write("Sending Tasks to celery for processing!")
                update_category.delay(c)



from ninja import NinjaAPI

from corptools.api.helpers import get_alts_queryset, get_main_character
from corptools.models.eve_models import EveItemDogmaAttribute, EveItemType
from corptools.models.skills import SkillList
from corptools.task_helpers.skill_helpers import SkillListCache

from ... import models


class FittingsApiEndpoints:
    tags = ["Extras"]

    def parse_eft_to_items(self, input: str):
        items = []
        for line in input.splitlines():
            _l = line.strip()
            if not len(_l):
                continue
            if _l.startswith("[") and _l.endswith("]"):
                # Header block
                if "," in _l:
                    _p = _l[1:-1].split(",")
                    items.append(_p[0].strip())
            else:
                _q = _l.split()[-1]
                if _q.startswith("x"):
                    # has a qty ignore and add the item
                    if "," in _l.split(_q)[0]:
                        # ammo in gun?
                        _i = _l.split(_q).split(",")
                        for i in _i:
                            items.append(i.strip())
                    else:
                        # any item in cargo etc
                        items.append(_l.split(_q)[0].strip())
                else:
                    if "," in _l:
                        _i = _l.split(",")
                        for i in _i:
                            items.append(i.strip())
                    else:
                        items.append(_l.strip())
        return items

    def __init__(self, api: NinjaAPI):
        @api.get(
            "/extras/fit2skills/{fit_id}",
            response={200: dict, 403: str, 404: str, 500: str},
            tags=["Fittings"]
        )
        def get_fit_skills(request, fit_id: str):
            """
                Turn a Fitting into a skill list json.
            """
            if not request.user.is_superuser:
                return 403, {"message": "Hard no pall!"}
            try:
                from fittings.models import Fitting, FittingItem
            except ImportError:
                return 500, "Fittings module not found!"
            try:
                _fit = Fitting.objects.get(
                    id=int(request.resolver_match.kwargs['fit_id']))
            except Fitting.DoesNotExist:
                return 404, "Fitting not found"
            _items = FittingItem.objects.filter(
                fit=_fit).values_list("type_id", flat=True)

            _skill_ids = [182, 183, 184, 1285, 1289, 1290]
            _level_ids = [277, 278, 279, 1286, 1287, 1288]

            _types = models.EveItemDogmaAttribute.objects.filter(
                eve_type_id__in=_items, attribute_id__in=_skill_ids + _level_ids)
            _types = _types | models.EveItemDogmaAttribute.objects.filter(
                eve_type_id=_fit.ship_type_type_id, attribute_id__in=_skill_ids + _level_ids)

            required = {}
            skills = {}
            skill_ids = set()
            for t in _types:
                if t.eve_type_id not in required:
                    required[t.eve_type_id] = {0: {"skill": 0, "level": 0},
                                               1: {"skill": 0, "level": 0},
                                               2: {"skill": 0, "level": 0},
                                               3: {"skill": 0, "level": 0},
                                               4: {"skill": 0, "level": 0},
                                               5: {"skill": 0, "level": 0}}
                a = t.attribute_id
                v = t.value
                if a in _skill_ids:
                    required[t.eve_type_id][_skill_ids.index(a)]["skill"] = v
                elif a in _level_ids:
                    indx = _level_ids.index(a)
                    if required[t.eve_type_id][indx]["level"] < v:
                        required[t.eve_type_id][indx]["level"] = v

                for t in required.values():
                    for s in t.values():
                        if s["skill"]:
                            if s["skill"] not in skills:
                                skills[s["skill"]] = {"skill": "", "level": 0}
                                skill_ids.add(s["skill"])
                            if s["level"] > skills[s["skill"]]["level"]:
                                skills[s["skill"]]["level"] = s["level"]
                out = {}
                for t in models.EveItemType.objects.filter(type_id__in=list(skill_ids)):
                    out[t.name] = skills[t.type_id]['level']

            return {
                "Fitting": _fit.name,
                "skill_list": out
            }

        @api.post(
            "/extras/fitting2skills/{character_id}",
            response={200: dict, 403: str, 404: str, 500: str},
            tags=["Fittings"],
            openapi_extra={
                "requestBody": {
                    "content": {
                        "text/plain": {
                            "schema": {
                                "type": "str",
                            }
                        }
                    },
                    "required": True,
                }
            },
        )
        def get_fitting_skills(request, character_id: int):
            """
                Turn a Fitting into a skill list and check characters on account.
            """
            if character_id == 0:
                character_id = request.user.profile.main_character.character_id
            response, main = get_main_character(request, character_id)
            if not response:
                return 403, _("Permission Denied")
            characters = get_alts_queryset(main)

            _in = str(request.body.decode())
            items = self.parse_eft_to_items(_in)
            _items = EveItemType.objects.filter(name__in=items)

            _skill_ids = [182, 183, 184, 1285, 1289, 1290]
            _level_ids = [277, 278, 279, 1286, 1287, 1288]

            _types = EveItemDogmaAttribute.objects.filter(
                eve_type_id__in=_items, attribute_id__in=_skill_ids + _level_ids)

            required = {}
            skills = {}
            sids = set()
            for t in _types:
                if t.eve_type_id not in required:
                    required[t.eve_type_id] = {0: {"skill": 0, "level": 0},
                                               1: {"skill": 0, "level": 0},
                                               2: {"skill": 0, "level": 0},
                                               3: {"skill": 0, "level": 0},
                                               4: {"skill": 0, "level": 0},
                                               5: {"skill": 0, "level": 0}}
                a = t.attribute_id
                v = t.value
                if a in _skill_ids:
                    required[t.eve_type_id][_skill_ids.index(a)]["skill"] = v
                elif a in _level_ids:
                    indx = _level_ids.index(a)
                    if required[t.eve_type_id][indx]["level"] < v:
                        required[t.eve_type_id][indx]["level"] = v

                for t in required.values():
                    for s in t.values():
                        if s["skill"]:
                            if s["skill"] not in skills:
                                skills[s["skill"]] = {
                                    "s": s["skill"], "l": 0, "n": ""}
                                sids.add(s["skill"])
                            if s["level"] > skills[s["skill"]]["l"]:
                                skills[s["skill"]]["l"] = s["level"]
            sk_check = {

            }
            for t in EveItemType.objects.filter(type_id__in=list(sids)):
                skills[t.type_id]["n"] = t.name
                sk_check[t.name] = skills[t.type_id]["l"]

            import json
            char_ids = list(
                characters.values_list('character_id', flat=True)
            )
            checks = SkillListCache().check_skill_lists(
                [
                    SkillList(
                        name="fit",
                        skill_list=json.dumps(sk_check)
                    )
                ],
                char_ids
            )
            # for c, d in checks.items():
            #     del (d["skills"])

            _resp = {
                "skills": list(sorted(skills.values(), key=lambda item: item["n"])),
                "chars": dict(sorted(checks.items()))
            }
            return _resp

        @api.get(
            "/extras/dogma/{type_id}",
            response={200: dict, 403: str, 404: str, 500: str},
            tags=["Fittings"]
        )
        def get_dogma(request, type_id: str):
            """
                Load dogma for a type_id.
            """
            if not request.user.is_superuser:
                return 403, {"message": "Hard no pall!"}

            _types = models.EveItemDogmaAttribute.objects.filter(
                eve_type_id=type_id)

            dogma = {}
            for t in _types:
                dogma[t.attribute_id] = t.value
            return dogma



# Generated by Django 4.2.16 on 2025-03-07 08:30

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('corptools', '0115_assetcoordiante'),
    ]

    operations = [
        migrations.CreateModel(
            name='Starbase',
            fields=[
                ('id', models.AutoField(auto_created=True,
                 primary_key=True, serialize=False, verbose_name='ID')),
                ('starbase_id', models.BigIntegerField()),
                ('onlined_since', models.DateTimeField(default=None, null=True)),
                ('reinforced_until', models.DateTimeField(default=None, null=True)),
                ('unanchor_at', models.DateTimeField(default=None, null=True)),
                ('allow_alliance_members', models.BooleanField(default=False)),
                ('allow_corporation_members', models.BooleanField(default=False)),
                ('attack_if_at_war', models.BooleanField(default=False)),
                ('attack_if_other_security_status_dropping',
                 models.BooleanField(default=False)),
                ('use_alliance_standings', models.BooleanField(default=False)),
                ('attack_security_status_threshold',
                 models.FloatField(default=None, null=True)),
                ('attack_standing_threshold',
                 models.FloatField(default=None, null=True)),
                ('state', models.CharField(choices=[('online', 'online'), ('offline', 'offline'), (
                    'onlining', 'onlining'), ('reinforced', 'reinforced'), ('unanchoring', 'unanchoring')], max_length=12)),
                ('anchor', models.CharField(choices=[('alliance_member', 'alliance_member'), ('config_starbase_equipment_role', 'config_starbase_equipment_role'), (
                    'corporation_member', 'corporation_member'), ('starbase_fuel_technician_role', 'starbase_fuel_technician_role')], default=None, max_length=31, null=True)),
                ('online', models.CharField(choices=[('alliance_member', 'alliance_member'), ('config_starbase_equipment_role', 'config_starbase_equipment_role'), (
                    'corporation_member', 'corporation_member'), ('starbase_fuel_technician_role', 'starbase_fuel_technician_role')], default=None, max_length=31, null=True)),
                ('offline', models.CharField(choices=[('alliance_member', 'alliance_member'), ('config_starbase_equipment_role', 'config_starbase_equipment_role'), (
                    'corporation_member', 'corporation_member'), ('starbase_fuel_technician_role', 'starbase_fuel_technician_role')], default=None, max_length=31, null=True)),
                ('unanchor', models.CharField(choices=[('alliance_member', 'alliance_member'), ('config_starbase_equipment_role', 'config_starbase_equipment_role'), (
                    'corporation_member', 'corporation_member'), ('starbase_fuel_technician_role', 'starbase_fuel_technician_role')], default=None, max_length=31, null=True)),
                ('fuel_bay_take', models.CharField(choices=[('alliance_member', 'alliance_member'), ('config_starbase_equipment_role', 'config_starbase_equipment_role'), (
                    'corporation_member', 'corporation_member'), ('starbase_fuel_technician_role', 'starbase_fuel_technician_role')], default=None, max_length=31, null=True)),
                ('fuel_bay_view', models.CharField(choices=[('alliance_member', 'alliance_member'), ('config_starbase_equipment_role', 'config_starbase_equipment_role'), (
                    'corporation_member', 'corporation_member'), ('starbase_fuel_technician_role', 'starbase_fuel_technician_role')], default=None, max_length=31, null=True)),
                ('fuels', models.TextField(blank=True, default='', null=True)),
                ('corporation', models.ForeignKey(
                    on_delete=django.db.models.deletion.CASCADE, to='corptools.corporationaudit')),
                ('moon', models.ForeignKey(default=None, null=True,
                 on_delete=django.db.models.deletion.CASCADE, to='corptools.mapsystemmoon')),
                ('system', models.ForeignKey(default=None, null=True,
                 on_delete=django.db.models.deletion.CASCADE, to='corptools.mapsystem')),
                ('type_name', models.ForeignKey(default=None, null=True,
                 on_delete=django.db.models.deletion.CASCADE, to='corptools.eveitemtype')),
            ],
        ),
    ]



# Generated by Django 4.2.13 on 2024-05-27 08:31

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('corptools', '0104_remove_currentshipfilter_categories_and_more'),
    ]

    operations = [
        migrations.CreateModel(
            name='HomeStationFilter',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=500)),
                ('description', models.CharField(max_length=500)),
                ('constellations', models.ManyToManyField(blank=True, help_text='Limit filter to specific Constellations', to='corptools.mapconstellation')),
                ('regions', models.ManyToManyField(blank=True, help_text='Limit filter to specific Regions', to='corptools.mapregion')),
                ('structures', models.ManyToManyField(blank=True, help_text='Limit filter to specific Structures', to='corptools.evelocation')),
                ('systems', models.ManyToManyField(blank=True, help_text='Limit filter to specific Systems', to='corptools.mapsystem')),
            ],
            options={
                'verbose_name': 'Smart Filter: Home Station (Death Clone)',
                'verbose_name_plural': 'Smart Filter: Home Station (Death Clone)',
            },
        ),
    ]



from ..models import CorporationAudit
from ..models.interactions import CharacterRoles
from . import CorptoolsTestCase


class TestCorptoolsCorpAccessPerms(CorptoolsTestCase):

    def test_no_perms_get_self_u1(self):
        cs = CorporationAudit.objects.visible_to(self.user1)
        self.assertNotIn(self.cp1, cs)
        self.assertNotIn(self.cp2, cs)
        self.assertNotIn(self.cp3, cs)
        self.assertNotIn(self.cp4, cs)

    def test_no_perms_get_self_u2(self):
        cs = CorporationAudit.objects.visible_to(self.user2)
        self.assertNotIn(self.cp1, cs)
        self.assertNotIn(self.cp2, cs)
        self.assertNotIn(self.cp3, cs)
        self.assertNotIn(self.cp4, cs)

    def test_no_perms_get_self_u3(self):
        cs = CorporationAudit.objects.visible_to(self.user3)
        self.assertNotIn(self.cp1, cs)
        self.assertNotIn(self.cp2, cs)
        self.assertNotIn(self.cp3, cs)
        self.assertNotIn(self.cp4, cs)

    def test_get_corp_u1(self):
        self.user1.user_permissions.add(self.own_corp_manager)
        self.user1.refresh_from_db()
        cs = CorporationAudit.objects.visible_to(self.user1)
        self.assertIn(self.cp1, cs)
        self.assertNotIn(self.cp2, cs)
        self.assertNotIn(self.cp3, cs)
        self.assertNotIn(self.cp4, cs)

    def test_get_corp_u2(self):
        self.user2.user_permissions.add(self.own_corp_manager)
        self.user2.refresh_from_db()
        cs = CorporationAudit.objects.visible_to(self.user2)
        self.assertNotIn(self.cp1, cs)
        self.assertIn(self.cp2, cs)
        self.assertNotIn(self.cp3, cs)
        self.assertNotIn(self.cp4, cs)

    def test_get_corp_u3(self):
        self.user3.user_permissions.add(self.own_corp_manager)
        self.user3.refresh_from_db()
        cs = CorporationAudit.objects.visible_to(self.user3)
        self.assertNotIn(self.cp1, cs)
        self.assertNotIn(self.cp2, cs)
        self.assertIn(self.cp3, cs)
        self.assertNotIn(self.cp4, cs)

    def test_get_alliance_u1(self):
        self.user1.user_permissions.add(self.alliance_corp_manager)
        self.user1.refresh_from_db()
        cs = CorporationAudit.objects.visible_to(self.user1)
        self.assertIn(self.cp1, cs)
        self.assertNotIn(self.cp2, cs)
        self.assertNotIn(self.cp3, cs)
        self.assertNotIn(self.cp4, cs)

    def test_get_alliance_u2(self):
        self.user2.user_permissions.add(self.alliance_corp_manager)
        self.user2.refresh_from_db()
        cs = CorporationAudit.objects.visible_to(self.user2)
        self.assertNotIn(self.cp1, cs)
        self.assertIn(self.cp2, cs)
        self.assertNotIn(self.cp3, cs)
        self.assertNotIn(self.cp4, cs)

    def test_get_alliance_u3(self):
        self.user3.user_permissions.add(self.alliance_corp_manager)
        self.user3.refresh_from_db()
        cs = CorporationAudit.objects.visible_to(self.user3)
        self.assertNotIn(self.cp1, cs)
        self.assertNotIn(self.cp2, cs)
        self.assertIn(self.cp3, cs)
        self.assertIn(self.cp4, cs)

    def test_global_perms_u1(self):
        self.user1.user_permissions.add(self.global_corp_manager)
        self.user1.refresh_from_db()
        cs = CorporationAudit.objects.visible_to(self.user1)
        self.assertIn(self.cp1, cs)
        self.assertIn(self.cp2, cs)
        self.assertIn(self.cp3, cs)
        self.assertIn(self.cp4, cs)

    def test_global_perms_u2(self):
        self.user2.user_permissions.add(self.global_corp_manager)
        self.user2.refresh_from_db()
        cs = CorporationAudit.objects.visible_to(self.user2)
        self.assertIn(self.cp1, cs)
        self.assertIn(self.cp2, cs)
        self.assertIn(self.cp3, cs)
        self.assertIn(self.cp4, cs)

    def test_global_perms_u3(self):
        self.user3.user_permissions.add(self.global_corp_manager)
        self.user3.refresh_from_db()
        cs = CorporationAudit.objects.visible_to(self.user3)
        self.assertIn(self.cp1, cs)
        self.assertIn(self.cp2, cs)
        self.assertIn(self.cp3, cs)
        self.assertIn(self.cp4, cs)

    def test_director_perms_u3(self):
        # own corp 3
        # alt corp 4
        self.user3.user_permissions.add(self.own_corp_manager)
        self.user3.user_permissions.add(self.director_manager)
        self.user3.refresh_from_db()
        cs = CorporationAudit.objects.visible_to(self.user3)
        self.assertNotIn(self.cp1, cs)
        self.assertNotIn(self.cp2, cs)
        self.assertIn(self.cp3, cs)
        self.assertIn(self.cp4, cs)

    def test_director_perms_after_corp_change_u3(self):
        # own corp 3
        # alt corp 4
        self.user3.user_permissions.add(self.own_corp_manager)
        self.user3.user_permissions.add(self.director_manager)
        self.user3.refresh_from_db()

        # swap alliance on char7 to corp 1
        self.char7.alliance_id = 109
        self.char7.alliance_name = "alliance.name1"
        self.char7.alliance_ticker = "ABC-d"
        self.char7.save()

        # No Changes
        cs = CorporationAudit.objects.visible_to(self.user3)
        self.assertNotIn(self.cp1, cs)
        self.assertNotIn(self.cp2, cs)
        self.assertIn(self.cp3, cs)
        self.assertIn(self.cp4, cs)

        # swap corps on char7 to corp 1
        self.char7.corporation_id = 123
        self.char7.corporation_name = "corporation.name1"
        self.char7.corporation_ticker = "ABC"
        self.char7.save()

        # refresh the char
        self.user3.refresh_from_db()

        # prove no leaks in perms
        cs = CorporationAudit.objects.visible_to(self.user3)
        self.assertNotIn(self.cp1, cs)
        self.assertNotIn(self.cp2, cs)
        self.assertIn(self.cp3, cs)
        self.assertNotIn(self.cp4, cs)

    def test_only_director_perms_u3(self):
        # own corp 3
        # alt corp 4
        self.user3.user_permissions.add(self.director_manager)
        self.user3.refresh_from_db()
        cs = CorporationAudit.objects.visible_to(self.user3)
        self.assertNotIn(self.cp1, cs)
        self.assertNotIn(self.cp2, cs)
        self.assertNotIn(self.cp3, cs)
        self.assertIn(self.cp4, cs)



# Generated by Django 4.0.10 on 2023-05-22 10:16

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('corptools', '0087_characteraudit_last_update_indy'),
    ]

    operations = [
        migrations.AlterField(
            model_name='characterindustryjob',
            name='end_date',
            field=models.DateTimeField(),
        ),
    ]



from typing import TYPE_CHECKING

from model_utils import Choices

from django.db import models
from django.utils.translation import gettext_lazy as _

from corptools.task_helpers import sanitize_notification_type

from .audits import CharacterAudit, CorporationAudit
from .eve_models import EveName

if TYPE_CHECKING:
    from esi.stubs import (
        CharactersCharacterIdContactsGetItem,
        CharactersCharacterIdContactsLabelsGetItem,
    )


class NotificationText(models.Model):
    notification_id = models.BigIntegerField(primary_key=True)
    notification_text = models.TextField(null=True, default=None)


class Notification(models.Model):
    id = models.BigAutoField(primary_key=True)
    character = models.ForeignKey(CharacterAudit, on_delete=models.CASCADE)

    notification_id = models.BigIntegerField()
    sender_id = models.IntegerField()
    _type_enum = Choices('character', 'corporation',
                         'alliance', 'faction', 'other')
    sender_type = models.CharField(max_length=15, choices=_type_enum)
    timestamp = models.DateTimeField()
    notification_type = models.CharField(max_length=50)
    is_read = models.BooleanField(null=True, default=None)
    notification_text = models.ForeignKey(
        NotificationText, on_delete=models.CASCADE, blank=True, null=True, default=None)

    class Meta:
        indexes = (
            models.Index(fields=['timestamp']),
            models.Index(fields=['notification_type'])
        )

    @classmethod
    def from_esi_model(cls, character, model):
        return (
            cls(
                character=character,
                notification_id=model.notification_id,
                sender_id=model.sender_id,
                sender_type=model.sender_type,
                notification_text_id=model.notification_id,
                timestamp=model.timestamp,
                notification_type=sanitize_notification_type(
                    model.type
                ),
                is_read=model.is_read
            ),
            NotificationText(
                notification_id=model.notification_id,
                notification_text=model.text
            )
        )


class MailLabel(models.Model):
    character = models.ForeignKey(CharacterAudit, on_delete=models.CASCADE)
    label_id = models.IntegerField(default=None)
    label_name = models.CharField(max_length=255, null=True, default=None)
    label_color = models.CharField(max_length=7, null=True, default=None)
    unread_count = models.IntegerField(null=True, default=None)


class MailRecipient(models.Model):
    recipient_id = models.BigIntegerField(primary_key=True, unique=True)
    recipient_name = models.ForeignKey(
        EveName, on_delete=models.SET_NULL, null=True, default=None)
    _recipient_enum = Choices('alliance', 'character',
                              'corporation', 'mailing_list')
    recipient_type = models.CharField(max_length=15, choices=_recipient_enum)


class MailMessage(models.Model):
    id_key = models.BigIntegerField(primary_key=True, unique=True)
    character = models.ForeignKey(CharacterAudit, on_delete=models.CASCADE)

    # headers
    mail_id = models.IntegerField(null=True, default=None)
    from_id = models.IntegerField(null=True, default=None)
    from_name = models.ForeignKey(
        EveName, on_delete=models.SET_NULL, null=True, default=None)

    recipients = models.ManyToManyField(MailRecipient)
    labels = models.ManyToManyField(MailLabel)
    is_read = models.BooleanField(null=True, default=False)
    timestamp = models.DateTimeField(null=True, default=None)

    # message
    subject = models.CharField(max_length=255, null=True, default=None)

    # not in bulk data
    body = models.CharField(max_length=12000, null=True, default=None)


class Contact(models.Model):
    id = models.BigIntegerField(primary_key=True)
    contact_id = models.BigIntegerField()
    contact_type = models.CharField(max_length=255, null=False)
    contact_name = models.ForeignKey(EveName, on_delete=models.CASCADE)
    standing = models.DecimalField(max_digits=4, decimal_places=2)

    class Meta:
        abstract = True


class ContactLabel(models.Model):
    id = models.BigIntegerField(primary_key=True)
    label_id = models.BigIntegerField()
    label_name = models.CharField(max_length=255, null=False)

    class Meta:
        abstract = True


class CharacterContactLabel(ContactLabel):
    character = models.ForeignKey(CharacterAudit, on_delete=models.CASCADE)

    def build_id(self):
        self.id = int(str(self.character_id) + str(self.label_id))
        return self.id

    @classmethod
    def from_esi_model(cls, character: CharacterAudit, esi_model: "CharactersCharacterIdContactsLabelsGetItem"):
        _mdl = cls(
            character=character,
            label_id=esi_model.label_id,
            label_name=esi_model.label_name,
        )
        _mdl.build_id()
        return _mdl


class CorporationContactLabel(ContactLabel):
    corporation = models.ForeignKey(CorporationAudit, on_delete=models.CASCADE)


class CharacterContact(Contact):
    character = models.ForeignKey(CharacterAudit, on_delete=models.CASCADE)
    blocked = models.BooleanField(default=False)
    watched = models.BooleanField(default=False)
    labels = models.ManyToManyField(CharacterContactLabel)

    def build_id(self):
        self.id = int(str(self.character_id) + str(self.contact_id))
        return self.id

    @classmethod
    def from_esi_model(cls, character: CharacterAudit, esi_model: "CharactersCharacterIdContactsGetItem"):
        blocked = False if esi_model.is_blocked is None else esi_model.is_blocked
        watched = False if esi_model.is_watched is None else esi_model.is_watched
        _mdl = cls(
            character=character,
            contact_id=esi_model.contact_id,
            contact_type=esi_model.contact_type,
            contact_name_id=esi_model.contact_id,
            standing=esi_model.standing,
            blocked=blocked,
            watched=watched,
        )
        _mdl.build_id()
        return _mdl


class CorporationContact(Contact):
    corporation = models.ForeignKey(CorporationAudit, on_delete=models.CASCADE)
    watched = models.BooleanField(default=False)
    labels = models.ManyToManyField(CorporationContactLabel)


class LoyaltyPoint(models.Model):
    character = models.ForeignKey(
        CharacterAudit, verbose_name=_("Character"), on_delete=models.CASCADE)
    corporation = models.ForeignKey(
        EveName, verbose_name=_("NPC Corporation"), on_delete=models.CASCADE)
    amount = models.IntegerField(_("LP"))

    class Meta:
        verbose_name = _("Loyalty Point")
        unique_together = ('character', 'corporation',)

    def __str__(self):
        return f"{self.character} - {self.corporation}"


class CorporationHistory(models.Model):
    character = models.ForeignKey(CharacterAudit, on_delete=models.CASCADE)

    corporation_id = models.IntegerField()
    corporation_name = models.ForeignKey(EveName, on_delete=models.CASCADE)
    is_deleted = models.BooleanField(null=True, default=None)
    record_id = models.IntegerField()
    start_date = models.DateTimeField()


class CharacterTitle(models.Model):
    title_id = models.IntegerField()
    title = models.CharField(max_length=500)
    corporation_id = models.BigIntegerField()
    corporation_name = models.CharField(max_length=500)

    def __str__(self):
        return f"({self.corporation_name}) - {self.title}"


class CharacterRoles(models.Model):
    character = models.OneToOneField(CharacterAudit, on_delete=models.CASCADE)

    director = models.BooleanField(default=False)
    accountant = models.BooleanField(default=False)
    station_manager = models.BooleanField(default=False)
    personnel_manager = models.BooleanField(default=False)

    titles = models.ManyToManyField(CharacterTitle)



import glob
import logging
import os
import shutil
import zipfile

import httpx

from ..models.eve_models import (
    EveItemCategory, EveItemDogmaAttribute, EveItemGroup, EveItemType,
    InvTypeMaterials, MapConstellation, MapRegion, MapSystem, MapSystemGate,
    MapSystemMoon, MapSystemPlanet,
)

logger = logging.getLogger(__name__)

# What models and the order to load them
SDE_PARTS_TO_UPDATE = [
    # Map
    MapRegion,
    MapConstellation,
    MapSystem,
    # System stuffs
    MapSystemPlanet,
    MapSystemMoon,
    MapSystemGate,
    # Types
    EveItemCategory,
    EveItemGroup,
    EveItemType,
    EveItemDogmaAttribute,
    # Type Materials
    InvTypeMaterials,
]

SDU_URL = "https://developers.eveonline.com/static-data/eve-online-static-data-latest-jsonl.zip"
SDE_FILE_NAME = "eve-online-static-data-latest-jsonl.zip"
SDE_FOLDER = "eve-sde"


def download_file(url, local_filename):
    """
    Downloads a file from a given URL using httpx and saves it locally.

    Args:
        url (str): The URL of the file to download.
        local_filename (str): The path and name to save the downloaded file.
    """
    try:
        with httpx.stream("GET", url, follow_redirects=True) as response:
            response.raise_for_status()  # Raise an exception for HTTP errors (4xx or 5xx)
            with open(local_filename, "wb") as f:
                for chunk in response.iter_bytes():
                    f.write(chunk)
        print(f"File downloaded successfully to: {local_filename}")
    except httpx.HTTPStatusError as e:
        print(f"HTTP error during download: {e}")
    except httpx.RequestError as e:
        print(f"Network error during download: {e}")
    except Exception as e:
        print(f"An unexpected error occurred: {e}")


def delete_sde_zip():
    os.remove(SDE_FILE_NAME)


def delete_sde_folder():
    shutil.rmtree(SDE_FOLDER)


def download_extract_sde():
    download_file(
        SDU_URL,
        SDE_FILE_NAME
    )
    with zipfile.ZipFile(SDE_FILE_NAME, mode="r") as zf:
        zf.extractall(path=SDE_FOLDER)
    # delete the zip
    delete_sde_zip()


def process_section_of_sde(id: int = 0):
    """
        Update a SDE model.
    """
    SDE_PARTS_TO_UPDATE[id].load_from_sde(SDE_FOLDER)


def process_from_sde(start_from: int = 0):
    """
        Update the SDE models in order.
    """
    download_extract_sde()

    count = 0
    for mdl in SDE_PARTS_TO_UPDATE:
        if count >= start_from:
            logger.info(f"Starting {mdl}")
            process_section_of_sde(count)
        else:
            logger.info(f"Skipping {mdl}")
        count += 1

    delete_sde_folder()



import logging

from django.db.models.signals import pre_save
from django.dispatch import receiver

from allianceauth.eveonline.models import EveCharacter

logger = logging.getLogger(__name__)


@receiver(pre_save, sender=EveCharacter)
def my_pre_save_handler(sender, instance, **kwargs):
    # need to invalidate the roles on character corp swaps
    # we can pull new data on next refresh when the esi give it to us.
    # this will not fix issues were the data is in cache after a corp swap happened

    try:
        if hasattr(instance, "characteraudit"):
            ca = instance.characteraudit
            if hasattr(ca, "characterroles"):
                roles = ca.characterroles
                old = EveCharacter.objects.get(
                    pk=instance.pk
                )
                if old.corporation_id != instance.corporation_id:
                    roles.director = False
                    roles.accountant = False
                    roles.station_manager = False
                    roles.personnel_manager = False
                    roles.titles.clear()
                    roles.save()
    except Exception as e:
        logger.error(e)



import logging
from typing import List

from ninja import NinjaAPI

from django.db.models import (
    ExpressionWrapper, FloatField, OuterRef, Subquery, Sum,
)
from django.utils import timezone

from allianceauth.services.hooks import get_extension_logger

from corptools import models
from corptools.api import schema
from corptools.task_helpers.update_tasks import fetch_location_name

logger = get_extension_logger(__name__)


class DashboardApiEndpoints:

    tags = ["Dashboards"]

    def __init__(self, api: NinjaAPI):
        @api.get(
            "dashboard/gates",
            response={200: List, 403: schema.Message},
            tags=self.tags
        )
        def get_dashboard_gates(request):
            perms = (
                request.user.has_perm('corptools.own_corp_manager')
                | request.user.has_perm('corptools.alliance_corp_manager')
                | request.user.has_perm('corptools.state_corp_manager')
                | request.user.has_perm('corptools.global_corp_manager')
                | request.user.has_perm('corptools.holding_corp_structures')
            )

            if not perms:
                logging.error(
                    f"Permission Denied for {request.user} to view structures!")
                return 403, {"message": "Permission Denied!"}

            output = []
            structures = models.Structure.get_visible(request.user).select_related(
                "corporation__corporation", "system_name"
            ).prefetch_related('structureservice_set').filter(type_id=35841)

            ozone = models.CorpAsset.objects.filter(
                type_id=16273,
                location_flag="StructureFuel"
            ).values("location_id").annotate(total=Sum('quantity'))
            levels = {}
            for o in ozone:
                if o["location_id"] not in levels:
                    levels[o["location_id"]] = 0
                levels[o["location_id"]] += o["total"]

            second_systems = set()
            output = {}
            now = timezone.now()
            for s in structures:
                split = s.name.split(" » ")
                from_sys = split[0]
                to_sys = split[1].split(" - ")[0]
                connection = f"{from_sys}{to_sys}"
                reverse_connection = f"{to_sys}{from_sys}"
                days = 0
                if s.fuel_expires:
                    days = (s.fuel_expires - now).days
                active = False
                for ss in s.structureservice_set.all():
                    if ss.name == "Jump Bridge Access" and ss.state == "online":
                        active = True
                if reverse_connection in second_systems:
                    output[to_sys]["end"] = {
                        "system_name": s.system_name.name,
                        "system_id": s.system_name_id,
                        "ozone": levels.get(s.structure_id),
                        "known": True,
                        "active": active,
                        "expires": days,
                        "name": s.name
                    }
                else:
                    output[from_sys] = {}
                    output[from_sys]["start"] = {
                        "system_name": s.system_name.name,
                        "system_id": s.system_name_id,
                        "ozone": levels.get(s.structure_id),
                        "known": True,
                        "active": active,
                        "expires": days,
                        "name": s.name
                    }
                    output[from_sys]["end"] = {
                        "known": False, "active": False}
                    second_systems.add(connection)

            return list(output.values())

        @api.get(
            "dashboard/sov",
            response={200: List, 403: str},
            tags=self.tags
        )
        def get_dashboard_sov(request):
            perms = (
                request.user.has_perm('corptools.holding_corp_assets')
            )

            if not perms:
                logging.error(
                    f"Permission Denied for {request.user} to view Sov Structures!")
                return 403, "Permission Denied!"

            types = [32458]

            assets = models.CorpAsset.get_visible(request.user).filter(
                type_id__in=types,
                location_type="solar_system").select_related(
                "type_name",
                "location_name__system",
                "location_name__system__constellation",
                "location_name__system__constellation__region",
                "type_name__group__category"
            )

            asset_locations = models.CorpAsset.get_visible(request.user).filter(
                location_id__in=assets.values("item_id")).select_related(
                "type_name"
            )

            location_names = {}

            for a in assets:
                if not a.location_name_id:
                    location = fetch_location_name(
                        a.location_id, a.location_type, 0)
                    a.location_name = location
                loc_id = a.item_id
                if loc_id not in location_names:
                    location_names[loc_id] = {
                        "system": {
                            "name": a.location_name.location_name,
                            "const": a.location_name.system.constellation.name,
                            "rgn": a.location_name.system.constellation.region.name
                        },
                        "upgrades": []
                    }

            for a in asset_locations:
                location_names[a.location_id]["upgrades"].append({
                    "id": a.type_name.type_id,
                    "name": a.type_name.name,
                    "active": a.location_flag
                })

            return list(location_names.values())

        @api.get(
            "dashboard/metenox",
            response={200: List[schema.Metenox], 403: str},
            tags=self.tags
        )
        def get_dashboard_drills(request):
            return self.get_dashboard_drills_levels(request)

        @api.get(
            "dashboard/dens",
            response={200: list, 403: str},
            tags=self.tags
        )
        def get_visible_dens(
            request,
        ):
            perms = (
                request.user.has_perm('corptools.own_corp_manager')
                | request.user.has_perm('corptools.alliance_corp_manager')
                | request.user.has_perm('corptools.state_corp_manager')
                | request.user.has_perm('corptools.global_corp_manager')
            )

            if not perms:
                logger.error(
                    f"Permission Denied for {request.user} to view mining!")
                return 403, "Permission Denied!"

            characters = models.CharacterAudit.objects.visible_to(
                request.user
            )
            assets = models.CharacterAsset.objects.filter(
                singleton=True,
                location_id__lte=32000000,
                type_name__group_id=4810,
                character__in=characters,
            ).select_related(
                "character",
                "character__character",
                "type_name",
                "location_name",
                "type_name__group__category"
            )

            output = []

            for a in assets:
                type_nm = a.type_name.name
                if a.name:
                    type_nm = f"{a.type_name.name} ({a.name})"
                loc = f"{a.location_id} ({a.location_flag})"
                const = {"id": 0, "name": ""}
                region = {"id": 0, "name": ""}
                if a.location_name:
                    loc = a.location_name.location_name
                    try:
                        const = {
                            "id": a.location_name.system.constellation_id,
                            "name": a.location_name.system.constellation.name
                        }
                        region = {
                            "id": a.location_name.system.constellation.region_id,
                            "name": a.location_name.system.constellation.region.name
                        }
                    except Exception:
                        pass
                character_name = a.character.character.character_name
                try:
                    mc = a.character.character.character_ownership.user.profile.main_character
                    character_name = f"{mc.character_name} ({character_name})"
                except Exception:
                    pass
                output.append({
                    "character": {
                        "character_id": a.character.character.character_id,
                        "character_name": character_name,
                        "corporation_id": a.character.character.corporation_id,
                        "corporation_name": a.character.character.corporation_name,
                        "alliance_id": a.character.character.alliance_id,
                        "alliance_name": a.character.character.alliance_name
                    },
                    "item": {
                        "id": a.type_name.type_id,
                        "name": type_nm,
                        "cat": f"{a.type_name.group.category.name} - {a.type_name.group.name}",
                        "cat_id": a.type_name.group.category_id
                    },
                    "quantity": a.quantity,
                    "id": a.item_id,
                    "expand": False,
                    "name": a.name,
                    "location": {
                        "id": a.location_id,
                        "name": loc,
                        "constellation": const,
                        "region": region
                    }
                })

            return output

    def get_dashboard_drills_levels(request):
        perms = (
            request.user.has_perm('corptools.own_corp_manager')
            | request.user.has_perm('corptools.alliance_corp_manager')
            | request.user.has_perm('corptools.state_corp_manager')
            | request.user.has_perm('corptools.global_corp_manager')
            | request.user.has_perm('corptools.holding_corp_structures')
        )

        if not perms:
            logging.error(
                f"Permission Denied for {request.user} to view Metenox!")
            return 403, "Permission Denied!"

        types = [81826]  # metenox drills

        drills = models.Structure.get_visible(request.user).filter(
            type_id__in=types
        ).select_related(
            "type_name",
            "system_name",
            "system_name__constellation",
            "system_name__constellation__region",
        )

        output = {}

        for d in drills:
            output[d.structure_id] = {
                "structure": {
                    "id": d.structure_id,
                    "owner": d.corporation.corporation,
                    "name": d.name,
                    "type": {
                        "id": d.type_name.type_id,
                        "name": d.type_name.name
                    },
                    "location": {
                        "id": d.system_name.system_id,
                        "name": d.system_name.name
                    },
                    "constellation": {
                        "id": d.system_name.constellation.constellation_id,
                        "name": d.system_name.constellation.name
                    },
                    "region": {
                        "id": d.system_name.constellation.region.region_id,
                        "name": d.system_name.constellation.region.name
                    },
                },
                "contents": {},
                "total": 0
            }

        type_price = models.TypePrice.objects.filter(
            item_id=OuterRef('type_id')
        )

        asset_contents = models.CorpAsset.objects.filter(
            location_id__in=drills.values("structure_id"),
            location_flag="MoonMaterialBay"
        ).select_related(
            "type_name"
        ).annotate(
            value=ExpressionWrapper(
                Subquery(type_price.values('price')) * Sum('quantity'),
                output_field=FloatField()
            )
        )

        for a in asset_contents:
            _lid = a.location_id
            _tid = a.type_id

            if _tid not in output[_lid]["contents"]:
                output[_lid]["contents"][_tid] = {
                    "type": {
                        "id": a.type_name.type_id,
                        "name": a.type_name.name
                    },
                    "quantity": 0,
                    "value": 0
                }

            output[_lid]["contents"][_tid]["quantity"] += a.quantity or 0
            output[_lid]["contents"][_tid]["value"] += a.value or 0
            output[_lid]["total"] += a.value or 0

        for k, i in output.items():
            i["contents"] = list(i["contents"].values())
            i["total"] = int(i["total"])

        return list(output.values())



import sys
import time
from datetime import timedelta

from celery import shared_task

from django.db.models import F
from django.db.models.functions import Power, Sqrt
from django.utils import timezone
from django.utils.html import strip_tags

from allianceauth.services.hooks import get_extension_logger
from esi.exceptions import HTTPNotModified
from esi.models import Token

from corptools.task_helpers.update_tasks import (
    fetch_location_name, load_system,
)

from .. import providers
from ..models import (
    CharacterAsset, CharacterAudit, CharacterContact, CharacterContactLabel,
    CharacterIndustryJob, CharacterLocation, CharacterMarketOrder,
    CharacterMiningLedger, CharacterRoles, CharacterTitle,
    CharacterWalletJournalEntry, CharAssetCoordiante, Clone, Contract,
    ContractItem, CorporationHistory, EveItemType, EveLocation, EveName,
    Implant, JumpClone, LoyaltyPoint, MailLabel, MailMessage, MailRecipient,
    MapSystemPlanet, Notification, NotificationText, Skill, SkillQueue,
    SkillTotalHistory, SkillTotals,
)

logger = get_extension_logger(__name__)


# def chunks(lst, n):
#     """Yield successive n-sized chunks from lst."""
#     for i in range(0, len(lst), n):
#         yield lst[i:i + n]


def get_token(character_id: int, scopes: list) -> "Token":
    """Helper method to get a valid token for a specific character with specific scopes.

    Args:
        character_id: Character to filter on.
        scopes: array of ESI scope strings to search for.

    Returns:
        Matching token or `False` when token is not found

    TODO: Push upstream to django-esi
    """
    token = (
        Token.objects
        .filter(character_id=character_id)
        .require_scopes(scopes)
        .require_valid()
        .first()
    )
    if token:
        return token
    else:
        return False


def update_character_location(character_id, force_refresh=False):
    audit_char = CharacterAudit.objects.get(
        character__character_id=character_id)
    logger.debug("updating location for: {}".format(
        audit_char.character.character_name))

    req_scopes = ['esi-location.read_ship_type.v1',
                  'esi-location.read_location.v1']

    token = get_token(character_id, req_scopes)

    if not token:
        return "No Tokens"

    try:
        location = providers.esi_openapi.char_location(
            character_id,
            token
        )
        _st = time.perf_counter()

        loc_id = None

        if location.structure_id:
            loc_id = location.structure_id
        elif location.station_id:
            loc_id = location.station_id
        else:
            loc_id = location.solar_system_id

        _loc = fetch_location_name(loc_id, "solar_system", character_id)

        if _loc:
            _loc.save()

        CharacterLocation.objects.update_or_create(
            character=audit_char,
            defaults={
                "current_location": _loc if _loc else None
            }
        )
        logger.debug(
            f"CT_TIME: {time.perf_counter() - _st} update_character_location_location {character_id}"
        )

    except HTTPNotModified:
        logger.info(
            f"CT: No New Location data for: {audit_char.character.character_name}"
        )
        pass

    try:
        ship_data = providers.esi_openapi.client.Location.GetCharactersCharacterIdShip(
            character_id=character_id,
            token=token
        ).result(
            force_refresh=force_refresh,
        )
        _st = time.perf_counter()

        ship, _ = EveItemType.objects.get_or_create_from_esi(
            ship_data.ship_type_id)

        CharacterLocation.objects.update_or_create(
            character=audit_char,
            defaults={
                "current_ship": ship,
                "current_ship_name": ship_data.ship_name,
            }
        )
        logger.debug(
            f"CT_TIME: {time.perf_counter() - _st} update_character_location_ship {character_id}"
        )

    except HTTPNotModified:
        logger.info("CT: No New current ship data for: {}".format(
            audit_char.character.character_name))
        pass

    try:
        req_scopes = ['esi-location.read_online.v1']

        token = get_token(character_id, req_scopes)

        if token:
            online_data = providers.esi_openapi.client.Location.GetCharactersCharacterIdOnline(
                character_id=character_id,
                token=token
            ).result(
                force_refresh=force_refresh,
            )
            _st = time.perf_counter()

            if online_data.last_login:
                audit_char.last_known_login = online_data.last_login

            if online_data.last_logout:
                audit_char.last_known_logoff = online_data.last_logout

            if online_data.logins:
                audit_char.total_logins = online_data.logins

            logger.debug(
                f"CT_TIME: {time.perf_counter() - _st} update_character_location_last_online {character_id}"
            )

    except HTTPNotModified:
        logger.info(
            f"CT: No New online data for: {audit_char.character.character_name}"
        )

    audit_char.last_update_location = timezone.now()
    audit_char.save()
    audit_char.is_active()

    return f"CT: Finished location/ship/online for: {audit_char.character.character_name}"


def update_corp_history(character_id, force_refresh=False):
    audit_char = CharacterAudit.objects.get(
        character__character_id=character_id)
    logger.debug(
        f"updating corp history for: {audit_char.character.character_name}"
    )
    try:
        corp_history = providers.esi_openapi.client.Character.GetCharactersCharacterIdCorporationhistory(
            character_id=character_id,
        ).result(
            force_refresh=force_refresh,
        )
        _st = time.perf_counter()
        for corp in corp_history:
            corp_name, _ = EveName.objects.get_or_create_from_esi(
                corp.corporation_id
            )
            try:
                _, _ = CorporationHistory.objects.update_or_create(
                    character=audit_char,
                    record_id=corp.record_id,
                    defaults={
                        'corporation_id': corp.corporation_id,
                        'corporation_name': corp_name,
                        'is_deleted': corp.is_deleted if corp.is_deleted else False,
                        'start_date': corp.start_date
                    }
                )
            except CorporationHistory.MultipleObjectsReturned:
                logger.error(
                    (
                        f"CT Corp History Duplicate - {audit_char.character.character_name} -"
                        f" Bad Data: {corp.model_dump_json(indent=2)}"
                    )
                )
        logger.debug(
            f"CT_TIME: {time.perf_counter() - _st} update_corp_history {character_id}"
        )
    except HTTPNotModified:
        logger.info(
            f"CT: No New pub data for: {audit_char.character.character_name}"
        )

    audit_char.last_update_pub_data = timezone.now()
    audit_char.save()
    audit_char.is_active()

    return f"CT: Finished pub data for: {audit_char.character.character_name}"


def update_character_skill_list(character_id, force_refresh=False):
    audit_char = CharacterAudit.objects.get(
        character__character_id=character_id
    )
    logger.debug(
        f"Updating Skills for: {audit_char.character.character_name}"
    )

    req_scopes = ['esi-skills.read_skills.v1']

    token = get_token(character_id, req_scopes)

    if not token:
        return "No Tokens"
    try:
        skills = providers.esi_openapi.client.Skills.GetCharactersCharacterIdSkills(
            character_id=character_id,
            token=token
        ).result(
            force_refresh=force_refresh
        )

        # Delete current SkillList
        _st = time.perf_counter()
        Skill.objects.filter(character=audit_char).delete()

        SkillTotals.objects.update_or_create(
            character=audit_char,
            defaults={
                'total_sp': skills.total_sp,
                'unallocated_sp': skills.unallocated_sp if skills.unallocated_sp else 0
            }
        )

        SkillTotalHistory.objects.create(
            character=audit_char,
            total_sp=skills.total_sp,
            unallocated_sp=skills.unallocated_sp if skills.unallocated_sp else 0
        )

        _check_skills = []
        _create_skills = []
        for skill in skills.skills:
            _check_skills.append(skill.skill_id)
            _skill = Skill(
                character=audit_char,
                skill_id=skill.skill_id,
                skill_name_id=skill.skill_id,
                active_skill_level=skill.active_skill_level,
                skillpoints_in_skill=skill.skillpoints_in_skill,
                trained_skill_level=skill.trained_skill_level
            )
            _create_skills.append(_skill)

        EveItemType.objects.create_bulk_from_esi(_check_skills)
        Skill.objects.bulk_create(_create_skills)

        logger.debug(
            f"CT_TIME: {time.perf_counter() - _st} update_character_skill_list {character_id}"
        )

    except HTTPNotModified:
        logger.info(
            f"CT: No New skills for: {audit_char.character.character_name}"
        )

    audit_char.last_update_skills = timezone.now()
    audit_char.save()
    audit_char.is_active()

    return f"CT: Finished skills for: {audit_char.character.character_name}"


def update_character_skill_queue(character_id, force_refresh=False):
    audit_char = CharacterAudit.objects.get(
        character__character_id=character_id)
    logger.debug("Updating Skill Queue for: {}".format(
        audit_char.character.character_name))

    req_scopes = ['esi-skills.read_skillqueue.v1']

    token = get_token(character_id, req_scopes)

    if not token:
        return "No Tokens"
    try:
        queue = providers.esi_openapi.client.Skills.GetCharactersCharacterIdSkillqueue(
            character_id=character_id,
            token=token
        ).result(
            force_refresh=force_refresh,
        )

        _st = time.perf_counter()
        # Delete current SkillList
        SkillQueue.objects.filter(
            character=audit_char
        ).delete()

        items = []
        _check_skills = []
        for item in queue:
            _check_skills.append(item.skill_id)
            queue_item = SkillQueue(
                character=audit_char,
                finish_level=item.finished_level,
                queue_position=item.queue_position,
                skill_id=item.skill_id,
                skill_name_id=item.skill_id,
                finish_date=item.finish_date,
                level_end_sp=item.level_end_sp,
                level_start_sp=item.level_start_sp,
                start_date=item.start_date,
                training_start_sp=item.training_start_sp
            )
            items.append(queue_item)

        EveItemType.objects.create_bulk_from_esi(_check_skills)
        SkillQueue.objects.bulk_create(items)

        logger.debug(
            f"CT_TIME: {time.perf_counter() - _st} update_character_skill_queue {character_id}"
        )

    except HTTPNotModified:
        logger.info(
            "CT: No New skill queue for: {}".format(
                audit_char.character.character_name
            )
        )

    audit_char.last_update_skill_que = timezone.now()
    audit_char.save()
    audit_char.is_active()

    return f"CT: Finished skill queue for: {audit_char.character.character_name}"


def update_character_assets(character_id, force_refresh=False):
    audit_char = CharacterAudit.objects.get(
        character__character_id=character_id)
    logger.debug("Updating Assets for: {}".format(
        audit_char.character.character_name))

    req_scopes = ['esi-assets.read_assets.v1']

    token = get_token(character_id, req_scopes)

    if not token:
        return "No Tokens"
    try:
        assets = providers.esi_openapi.client.Assets.GetCharactersCharacterIdAssets(
            character_id=character_id,
            token=token
        ).results(
            force_refresh=force_refresh
        )
        logger.info(
            f"CT: New assets for: {audit_char.character.character_name}")

        _st = time.perf_counter()
        location_names = list(
            EveLocation.objects.all().values_list('location_id', flat=True))

        _current_type_ids = []
        item_ids = []
        items = []

        for item in assets:
            if item.type_id not in _current_type_ids:
                _current_type_ids.append(item.type_id)
            item_ids.append(item.item_id)

            asset_item = CharacterAsset.from_esi_model(audit_char, item)

            # attach location name
            if item.location_id in location_names:
                asset_item.location_name_id = item.location_id

            items.append(asset_item)

        # current ship doesn't show if in space sometimes
        ship = get_current_ship_location(
            character_id,
            force_refresh=force_refresh
        )
        if ship:
            logger.info(
                f"CT: New ship for: {audit_char.character.character_name}"
            )
            if ship.item_id not in item_ids:
                if ship.type_id not in _current_type_ids:
                    _current_type_ids.append(ship.type_id)

                if ship.location_id in location_names:
                    ship.location_name_id = ship.location_id

                items.append(ship)

        EveItemType.objects.create_bulk_from_esi(_current_type_ids)

        delete_query = CharacterAsset.objects.filter(
            character=audit_char
        )  # Flush Assets
        if delete_query.exists():
            # We now have some FKeys so slow it down...
            delete_query.delete()

        CharacterAsset.objects.bulk_create(items)

        logger.debug(
            f"CT_TIME: {time.perf_counter() - _st} update_character_assets {character_id}"
        )

        # Locate assets in space!
        update_asset_locations(character_id, force_refresh=force_refresh)
        try:
            # Get Asset Names!
            update_character_assets_names(character_id)
        except Exception as e:
            logger.error(e)
        try:
            # Get den planets!
            update_den_locations(character_id)
        except Exception as e:
            logger.error(e)

    except HTTPNotModified as e:
        _, _, tb = sys.exc_info()
        logger.info(
            f"CT: No New assets for: {audit_char.character.character_name} - ({tb.tb_lineno})"
        )

    audit_char.last_update_assets = timezone.now()
    audit_char.save()
    audit_char.is_active()

    return f"CT: Finished assets for: {audit_char.character.character_name}"


def chunks(qst, n):
    """Yield successive n-sized chunks from lst."""
    for i in range(0, qst.count(), n):
        yield qst[i:i + n]


def update_character_assets_names(character_id):
    """
        Uses PostCharactersCharacterIdAssetsNames
    """
    audit_char = CharacterAudit.objects.get(
        character__character_id=character_id
    )

    logger.debug(
        f"Updating Asset names for: {audit_char.character.character_name}")

    req_scopes = ['esi-assets.read_assets.v1']

    token = get_token(character_id, req_scopes)

    if not token:
        return "No Tokens"

    expandable_cats = [2, 6]

    asset_list = CharacterAsset.objects.filter(
        character=audit_char,
        type_name__group__category_id__in=expandable_cats,
        singleton=True
    ).order_by("pk")

    for subset in chunks(asset_list, 100):
        assets_names = providers.esi_openapi.client.Assets.PostCharactersCharacterIdAssetsNames(
            character_id=character_id,
            body=[i.item_id for i in subset],
            token=token
        ).results()

        id_list = {i.item_id: i.name for i in assets_names}

        for asset in subset:
            if asset.item_id in id_list:
                asset.name = id_list.get(asset.item_id)
                asset.save()


def update_den_locations(character_id, force_refresh=False):
    max_location_id = 35000000
    min_location_id = 30000000

    parents = CharacterAsset.objects.filter(
        character__character__character_id=character_id,
        singleton=True,
        type_name__group_id=4810,
        location_id__gte=min_location_id,
        location_id__lte=max_location_id
    )
    for parent in parents:
        load_system(parent.location_name.system.system_id)
        distance = MapSystemPlanet.objects.filter(
            system=parent.location_name.system
        ).annotate(
            distance=Sqrt(
                Power(
                    parent.coordinate.x - F("x"),
                    2
                ) + Power(
                    parent.coordinate.y - F("y"),
                    2
                ) + Power(
                    parent.coordinate.z - F("z"),
                    2
                )
            )
        ).order_by("distance").first()
        parent.name = distance.name
        parent.save()


def update_asset_locations(character_id, force_refresh=False):
    """
        Uses PostCharactersCharacterIdAssetsLocations
    """
    audit_char = CharacterAudit.objects.get(
        character__character_id=character_id
    )

    logger.debug(
        f"Updating Asset locations for: {audit_char.character.character_name}")

    categories = [
        23,  # Starbases
        65,  # Structures
        40,  # Sov
        22,  # Deployables
    ]

    # These must be in space
    max_location_id = 35000000
    min_location_id = 30000000

    assets = CharacterAsset.objects.filter(
        character=audit_char,
        type_name__group__category_id__in=categories,
        location_id__gte=min_location_id,
        location_id__lte=max_location_id
    )

    req_scopes = ['esi-assets.read_assets.v1']

    _token = get_token(character_id, req_scopes)

    if not _token or not assets.count():
        logger.info(
            f"CT: COORDS No Tokens or Assets for {audit_char.character}")
        return f"CT: COORDS No Tokens or Assets for {audit_char.character}"

    locations = []

    for ids in chunks(assets, 900):
        locations = providers.esi_openapi.client.Assets.PostCharactersCharacterIdAssetsLocations(
            character_id=audit_char.character.character_id,
            body=list(ids.values_list("item_id", flat=True)),
            token=_token
        ).result()

    # logger.info(locations)

    new_coords = {}
    for loc in locations:
        new_coords[loc.item_id] = loc.position

    new_models = []
    for a in assets:
        if a.item_id in new_coords:
            new_models.append(
                CharAssetCoordiante(
                    item=a,
                    x=new_coords[a.item_id].x,
                    y=new_coords[a.item_id].y,
                    z=new_coords[a.item_id].z
                )
            )

    CharAssetCoordiante.objects.bulk_create(new_models, ignore_conflicts=True)

    return f"CT: Finished asset locations for: {audit_char.character.character_name}"


def update_character_mining(character_id, force_refresh=False):
    audit_char = CharacterAudit.objects.get(
        character__character_id=character_id
    )
    logger.debug(f"Updating Mining for: {audit_char.character.character_name}")

    req_scopes = ['esi-industry.read_character_mining.v1']

    token = get_token(character_id, req_scopes)

    if not token:
        return "No Tokens"
    try:
        mining = providers.esi_openapi.client.Industry.GetCharactersCharacterIdMining(
            character_id=character_id,
            token=token
        ).result(
            force_refresh=force_refresh,
        )

        _st = time.perf_counter()
        existing_pks = set(
            CharacterMiningLedger.objects.filter(
                character=audit_char,
                date__gte=timezone.now() - timedelta(days=30)
            ).values_list(
                "id",
                flat=True
            )
        )

        type_ids = set()
        new_events = []
        old_events = []

        for event in mining:
            type_ids.add(event.type_id)
            pk = CharacterMiningLedger.create_primary_key(character_id, event)
            _e = CharacterMiningLedger(
                id=pk,
                character=audit_char,
                date=event.date,
                type_name_id=event.type_id,
                system_id=event.solar_system_id,
                quantity=event.quantity
            )
            if pk in existing_pks:
                old_events.append(_e)
            else:
                new_events.append(_e)

        EveItemType.objects.create_bulk_from_esi(list(type_ids))

        if len(new_events):
            CharacterMiningLedger.objects.bulk_create(
                new_events, ignore_conflicts=True)

        if len(old_events):
            CharacterMiningLedger.objects.bulk_update(
                old_events,
                fields=['quantity']
            )

        logger.debug(
            f"CT_TIME: {time.perf_counter() - _st} update_character_mining {character_id}"
        )

    except HTTPNotModified:
        logger.info(
            f"CT: No New Mining for: {audit_char.character.character_name}"
        )

    audit_char.last_update_mining = timezone.now()
    audit_char.save()
    audit_char.is_active()

    return f"CT: Finished Mining for: {audit_char.character.character_name}"


def update_character_industry_jobs(character_id, force_refresh=False):
    audit_char = CharacterAudit.objects.get(
        character__character_id=character_id
    )
    logger.debug(
        f"Updating Industry Jobs for: {audit_char.character.character_name}"
    )

    req_scopes = ['esi-industry.read_character_jobs.v1']

    token = get_token(character_id, req_scopes)

    if not token:
        return "No Tokens"
    try:
        jobs = providers.esi_openapi.client.Industry.GetCharactersCharacterIdIndustryJobs(
            character_id=character_id,
            token=token
        ).result(
            force_refresh=force_refresh,
        )

        _st = time.perf_counter()
        existing_pks = set(
            CharacterIndustryJob.objects.filter(
                character=audit_char
            ).values_list(
                "job_id",
                flat=True
            )
        )
        type_ids = set()
        new_events = []
        # old_events = []
        for event in jobs:
            type_ids.add(event.blueprint_type_id)
            if event.product_type_id:
                type_ids.add(event.product_type_id)

            if event.job_id in existing_pks:
                _m = CharacterIndustryJob.objects.get(
                    character=audit_char,
                    job_id=event.job_id
                )
                _m.completed_character_id = event.completed_character_id
                _m.completed_date = event.completed_date
                _m.end_date = event.end_date
                _m.pause_date = event.pause_date
                _m.status = event.status
                _m.successful_runs = event.successful_runs
                _m.save()
                continue

            _e = CharacterIndustryJob(
                character=audit_char,
                activity_id=event.activity_id,
                blueprint_id=event.blueprint_id,
                blueprint_location_id=event.blueprint_location_id,
                blueprint_type_id=event.blueprint_type_id,
                blueprint_type_name_id=event.blueprint_type_id,
                completed_character_id=event.completed_character_id,
                completed_date=event.completed_date,
                cost=event.cost,
                duration=event.duration,
                end_date=event.end_date,
                facility_id=event.facility_id,
                installer_id=event.installer_id,
                job_id=event.job_id,
                licensed_runs=event.licensed_runs,
                output_location_id=event.output_location_id,
                pause_date=event.pause_date,
                probability=event.probability,
                product_type_id=event.product_type_id,
                product_type_name_id=event.product_type_id,
                runs=event.runs,
                start_date=event.start_date,
                station_id=event.station_id,
                status=event.status,
                successful_runs=event.successful_runs
            )
            new_events.append(_e)

        EveItemType.objects.create_bulk_from_esi(list(type_ids))

        if len(new_events):
            CharacterIndustryJob.objects.bulk_create(
                new_events,
                ignore_conflicts=True
            )

        logger.debug(
            f"CT_TIME: {time.perf_counter() - _st} update_character_industry_jobs {character_id}"
        )

    except HTTPNotModified:
        logger.info(
            f"CT: No New Industry for: {audit_char.character.character_name}"
        )

    audit_char.last_update_indy = timezone.now()
    audit_char.save()
    audit_char.is_active()

    return f"CT: Finished Industry Jobs for: {audit_char.character.character_name}"


def get_current_ship_location(character_id, force_refresh=False):
    """
        Uses GetCharactersCharacterIdShip, GetCharactersCharacterIdLocation
    """
    audit_char = CharacterAudit.objects.get(
        character__character_id=character_id
    )

    req_scopes = [
        'esi-location.read_location.v1',
        'esi-location.read_ship_type.v1'
    ]

    token = get_token(character_id, req_scopes)

    if not token:
        return False
    try:
        # Todo decouple this
        ship = providers.esi_openapi.client.Location.GetCharactersCharacterIdShip(
            character_id=character_id,
            token=token
        ).result(
            force_refresh=force_refresh,
            use_etag=False
        )

        location = providers.esi_openapi.client.Location.GetCharactersCharacterIdLocation(
            character_id=character_id,
            token=token
        ).result(
            force_refresh=force_refresh,
            use_etag=False
        )

        return CharacterAsset.from_esi_location(audit_char, ship, location)
    except HTTPNotModified:
        return False


def update_character_wallet(character_id, force_refresh=False):
    audit_char = CharacterAudit.objects.get(
        character__character_id=character_id)
    logger.debug("Updating wallet transactions for: {}".format(
        audit_char.character.character_name))

    req_scopes = ['esi-wallet.read_character_wallet.v1']

    token = get_token(character_id, req_scopes)

    if not token:
        return "No Tokens"

    try:
        journal_items = providers.esi_openapi.client.Wallet.GetCharactersCharacterIdWalletJournal(
            character_id=character_id,
            token=token
        ).result(
            force_refresh=force_refresh,
        )

        _st = time.perf_counter()
        _current_journal = CharacterWalletJournalEntry.objects.filter(
            entry_id__in=[  # only fetch the ones we care about to reduce DB loading/ram
                e.id for e in journal_items
            ],
            character=audit_char
        ).values_list('entry_id', flat=True)

        _current_eve_ids = list(
            EveName.objects.all().values_list('eve_id', flat=True)
        )

        _new_names = []

        items = []
        for item in journal_items:
            if item.id not in _current_journal:
                if item.second_party_id not in _current_eve_ids:
                    _new_names.append(item.second_party_id)
                    _current_eve_ids.append(item.second_party_id)
                if item.first_party_id not in _current_eve_ids:
                    _new_names.append(item.first_party_id)
                    _current_eve_ids.append(item.first_party_id)

                items.append(
                    CharacterWalletJournalEntry.from_esi_model(
                        audit_char,
                        item
                    )
                )

        created_names = EveName.objects.create_bulk_from_esi(_new_names)

        wallet_ballance = providers.esi_openapi.client.Wallet.GetCharactersCharacterIdWallet(
            character_id=character_id,
            token=token
        ).result(
            use_etag=False
        )

        audit_char.balance = wallet_ballance
        audit_char.save()

        if created_names:
            CharacterWalletJournalEntry.objects.bulk_create(items)
        else:
            raise Exception("ESI Fail")
        logger.debug(
            f"CT_TIME: {time.perf_counter() - _st} update_character_wallet {character_id}")
    except HTTPNotModified:
        logger.info(
            f"CT: No New wallet data for: {audit_char.character.character_name}"
        )

    audit_char.last_update_wallet = timezone.now()
    audit_char.save()
    audit_char.is_active()

    return f"CT: Finished wallet transactions for: {audit_char.character.character_name}"


def update_character_transactions(character_id, force_refresh=False):
    audit_char = CharacterAudit.objects.get(
        character__character_id=character_id)
    logger.debug("Updating wallet transactions for: {}".format(
        audit_char.character.character_name))

    req_scopes = ['esi-wallet.read_character_wallet.v1']

    token = get_token(character_id, req_scopes)

    if not token:
        return "No Tokens"

    try:
        journal_items = providers.esi_openapi.client.Wallet.GetCharactersCharacterIdWalletTransactions(
            character_id=character_id,
            token=token
        ).result(
            use_etag=False
        )

        _st = time.perf_counter()

        _current_journal = CharacterWalletJournalEntry.objects.filter(
            character=audit_char,
            context_id_type="market_transaction_id",
            # Max items from ESI
            reason__exact=""
        ).values_list(
            'context_id',
            flat=True
        )[:2500]

        # _new_names = []

        # items = []
        for item in journal_items:
            if item.transaction_id in _current_journal:
                type_name, _ = EveItemType.objects.get_or_create_from_esi(
                    item.type_id
                )
                message = f"{item.quantity}x {type_name.name} @ {item.unit_price:,.2f} ISK"
                CharacterWalletJournalEntry.objects.filter(
                    character=audit_char,
                    context_id_type="market_transaction_id",
                    reason__exact="",
                    context_id=item.transaction_id
                ).update(
                    reason=message
                )
        logger.debug(
            f"CT_TIME: {time.perf_counter() - _st} update_character_transactions {character_id}")

    except HTTPNotModified:
        logger.info(
            f"CT: No New wallet data for: {audit_char.character.character_name}"
        )

    return f"CT: Finished market transactions for: {audit_char.character.character_name}"


def update_character_clones(character_id, force_refresh=False):
    audit_char = CharacterAudit.objects.get(
        character__character_id=character_id)
    logger.debug("Updating Clones and Implants for: {}".format(
        audit_char.character.character_name))

    req_scopes = ['esi-clones.read_clones.v1', 'esi-clones.read_implants.v1']

    token = get_token(character_id, req_scopes)

    if not token:
        return "No Tokens"

    try:
        jump_clones = providers.esi_openapi.client.Clones.GetCharactersCharacterIdClones(
            character_id=character_id,
            token=token
        ).result()

        active_clone = providers.esi_openapi.client.Clones.GetCharactersCharacterIdImplants(
            character_id=character_id,
            token=token
        ).result(
            use_etag=False
        )

        all_locations = list(EveLocation.objects.all(
        ).values_list('location_id', flat=True))

        home_loc = None  # Setting this to  none will force a lookup later on.
        try:
            home_loc = EveLocation.objects.get(
                location_id=jump_clones.home_location.location_id
            )
        except EveLocation.DoesNotExist:
            pass

        char_clone, created = Clone.objects.update_or_create(
            character=audit_char,
            defaults={
                'last_clone_jump_date': jump_clones.last_clone_jump_date,
                'last_station_change_date': jump_clones.last_station_change_date,
                'location_id': jump_clones.home_location.location_id,
                'location_type': jump_clones.home_location.location_type,
                'location_name': home_loc
            }
        )

        JumpClone.objects.filter(character=audit_char).delete()  # remove all

        implants = []
        type_ids = []

        for clone in jump_clones.jump_clones:
            _jumpclone = JumpClone(
                character=audit_char,
                jump_clone_id=clone.jump_clone_id,
                location_id=clone.location_id,
                location_type=clone.location_type,
                name=clone.name
            )
            if clone.location_id in all_locations:
                _jumpclone.location_name_id = clone.location_id

            _jumpclone.save()

            for implant in clone.implants:
                if implant not in type_ids:
                    type_ids.append(implant)
                implants.append(
                    Implant(
                        clone=_jumpclone,
                        type_name_id=implant
                    )
                )

        _jumpclone = JumpClone.objects.create(
            character=audit_char,
            jump_clone_id=0,
            name="Active Clone"
        )

        for implant in active_clone:
            if implant not in type_ids:
                type_ids.append(implant)
            implants.append(
                Implant(
                    clone=_jumpclone,
                    type_name_id=implant,
                )
            )

        EveItemType.objects.create_bulk_from_esi(type_ids)
        Implant.objects.bulk_create(implants)
    except HTTPNotModified:
        logger.info(
            f"CT: No New Clone data for: {audit_char.character.character_name}"
        )

    audit_char.last_update_clones = timezone.now()
    audit_char.save()
    audit_char.is_active()

    return f"CT: Finished clones for: {audit_char.character.character_name}"


def update_character_loyaltypoints(character_id, force_refresh=False):
    audit_char = CharacterAudit.objects.get(
        character__character_id=character_id
    )
    logger.debug(
        f"Updating Loyalty Points for: {audit_char.character.character_name}"
    )

    req_scopes = ['esi-characters.read_loyalty.v1', ]

    token = get_token(character_id, req_scopes)

    if not token:
        return "No Tokens"

    existing_lp_corps = LoyaltyPoint.objects.filter(
        character=audit_char
    ).values_list("corporation_id", flat=True)

    try:
        loyaltypoints = providers.esi_openapi.client.Loyalty.GetCharactersCharacterIdLoyaltyPoints(
            character_id=character_id,
            token=token
        ).results(
            force_refresh=force_refresh
        )

        _bulkcreate = []
        _bulkupdate = []

        for lp in loyaltypoints:
            lp_corp, _ = EveName.objects.get_or_create_from_esi(
                lp.corporation_id
            )

            if lp.corporation_id in existing_lp_corps:
                existing = LoyaltyPoint.objects.get(
                    character=audit_char, corporation=lp_corp
                )
                existing.amount = lp.loyalty_points
                _bulkupdate.append(existing)
            else:
                _bulkcreate.append(
                    LoyaltyPoint(
                        character=audit_char,
                        corporation=lp_corp,
                        amount=lp.loyalty_points
                    )
                )

        LoyaltyPoint.objects.bulk_create(
            _bulkcreate,
            ignore_conflicts=True,
        )
        LoyaltyPoint.objects.bulk_update(
            _bulkupdate,
            fields=['amount']
        )

    except HTTPNotModified:
        logger.info(
            f"CT: No New LP data for: {audit_char.character.character_name}"
        )

    audit_char.last_update_loyaltypoints = timezone.now()
    audit_char.save()
    audit_char.is_active()

    return f"CT: Finished Loyalty Points for: {audit_char.character.character_name}"


def update_character_orders(character_id, force_refresh=False):
    audit_char = CharacterAudit.objects.get(
        character__character_id=character_id
    )
    logger.debug(
        f"Updating Market Orders for: {audit_char.character.character_name}"
    )

    req_scopes = ['esi-markets.read_character_orders.v1']

    token = get_token(character_id, req_scopes)

    if not token:
        return "No Tokens"

    try:
        open_orders = providers.esi_openapi.client.Market.GetCharactersCharacterIdOrders(
            character_id=character_id,
            token=token
        ).results(
            force_refresh=force_refresh
        )

        _st = time.perf_counter()

        open_ids = list(
            CharacterMarketOrder.objects.filter(
                character=audit_char,
                state='active'
            ).values_list("order_id", flat=True)
        )
        all_locations = list(
            EveLocation.objects.all().values_list('location_id', flat=True)
        )

        updates = []
        creates = []
        type_ids = []
        tracked_ids = []

        for order in open_orders:
            tracked_ids.append(order.order_id)

            if order.type_id not in type_ids:
                type_ids.append(order.type_id)

            _order = CharacterMarketOrder.from_esi_model(audit_char, order)

            if order.location_id in all_locations:
                _order.location_name_id = order.location_id

            if order.order_id in open_ids:
                updates.append(_order)
            else:
                creates.append(_order)

        EveItemType.objects.create_bulk_from_esi(type_ids)

        if len(updates) > 0:
            CharacterMarketOrder.objects.bulk_update(
                updates, fields=[
                    'duration',
                    'escrow',
                    'min_volume',
                    'price',
                    'order_range',
                    'volume_remain',
                    'volume_total',
                    'state'
                ]
            )

        if len(creates) > 0:
            CharacterMarketOrder.objects.bulk_create(creates)

        logger.debug(
            f"CT_TIME: {time.perf_counter() - _st} update_character_orders {character_id}"
        )

    except HTTPNotModified:
        logger.info(
            f"CT: No New orders data for: {audit_char.character.character_name}"
        )

    audit_char.last_update_orders = timezone.now()
    audit_char.save()
    audit_char.is_active()

    return f"CT: Finished Orders for: {audit_char.character.character_name}"


def update_character_order_history(character_id, force_refresh=False):
    audit_char = CharacterAudit.objects.get(
        character__character_id=character_id
    )
    logger.debug(
        f"Updating Market Order History for: {audit_char.character.character_name}"
    )

    req_scopes = ['esi-markets.read_character_orders.v1']

    token = get_token(character_id, req_scopes)

    if not token:
        return "No Tokens"

    try:
        order_history = providers.esi_openapi.client.Market.GetCharactersCharacterIdOrdersHistory(
            character_id=character_id,
            token=token
        ).results(
            force_refresh=force_refresh
        )
        _st = time.perf_counter()

        closed_ids = list(
            CharacterMarketOrder.objects.filter(
                character=audit_char
            ).values_list(
                "order_id",
                flat=True
            )
        )
        all_locations = list(
            EveLocation.objects.all().values_list(
                'location_id',
                flat=True
            )
        )

        updates = []
        creates = []
        type_ids = []

        tracked_ids = []

        for order in order_history:
            tracked_ids.append(order.order_id)

            if order.type_id not in type_ids:
                type_ids.append(order.type_id)

            _order = CharacterMarketOrder.from_esi_model(audit_char, order)

            if order.location_id in all_locations:
                _order.location_name_id = order.location_id

            if order.order_id in closed_ids:
                updates.append(_order)
            else:
                creates.append(_order)

        EveItemType.objects.create_bulk_from_esi(type_ids)

        if len(updates) > 0:
            CharacterMarketOrder.objects.bulk_update(
                updates,
                fields=[
                    'duration',
                    'escrow',
                    'min_volume',
                    'price',
                    'order_range',
                    'volume_remain',
                    'volume_total',
                    'state'
                ],
                batch_size=1000
            )

        if len(creates) > 0:
            CharacterMarketOrder.objects.bulk_create(
                creates,
                batch_size=1000
            )

        logger.debug(
            f"CT_TIME: {time.perf_counter() - _st} update_character_order_history {character_id}"
        )

    except HTTPNotModified:
        logger.info(
            f"CT: No New old orders data for: {audit_char.character.character_name}"
        )

    return f"CT: Finished Order History for: {audit_char.character.character_name}"


@shared_task
def update_character_notifications(character_id, force_refresh=False):
    """
        Uses GetCharactersCharacterIdNotifications
    """
    audit_char = CharacterAudit.objects.get(
        character__character_id=character_id
    )
    logger.debug(
        f"Updating Notifications for: {audit_char.character.character_name}"
    )

    req_scopes = ['esi-characters.read_notifications.v1']

    token = get_token(character_id, req_scopes)

    if not token:
        return "No Tokens"

    try:
        notifications = providers.esi_openapi.client.Character.GetCharactersCharacterIdNotifications(
            character_id=character_id,
            token=token,
        ).result(
            force_refresh=force_refresh
        )

        _st = time.perf_counter()

        last_five_hundred = list(
            Notification.objects.filter(
                character=audit_char
            ).order_by(
                '-timestamp'
            )[:500].values_list(
                'notification_id',
                flat=True
            )
        )

        _creates = []
        _create_notifs = []
        for note in notifications:
            if not note.notification_id in last_five_hundred:
                _note, _text = Notification.from_esi_model(audit_char, note)
                _creates.append(_note)
                _create_notifs.append(_text)

        NotificationText.objects.bulk_create(
            _create_notifs,
            ignore_conflicts=True,
            batch_size=500
        )
        Notification.objects.bulk_create(
            _creates,
            batch_size=500
        )

        logger.debug(
            f"CT_TIME: {time.perf_counter() - _st} update_character_notifications {character_id}"
        )

    except HTTPNotModified:
        logger.info(
            f"CT: No New notifications for: {audit_char.character.character_name}"
        )
        pass

    audit_char.last_update_notif = timezone.now()
    audit_char.save()
    audit_char.is_active()

    return f"CT: Finished notifications for: {audit_char.character.character_name}"


def update_character_roles(character_id, force_refresh=False):
    audit_char = CharacterAudit.objects.get(
        character__character_id=character_id)
    req_scopes = ['esi-characters.read_corporation_roles.v1']

    token = get_token(character_id, req_scopes)

    if not token:
        return False
    try:
        roles = providers.esi_openapi.client.Character.GetCharactersCharacterIdRoles(
            character_id=character_id,
            token=token,
        ).result(
            force_refresh=force_refresh
        )

        _st = time.perf_counter()

        director = False
        accountant = False
        station_manager = False
        personnel_manager = False

        if "Director" in roles.roles:
            director = True

        if "Accountant" in roles.roles:
            accountant = True

        if "Station_Manager" in roles.roles:
            station_manager = True

        if "Personnel_Manager" in roles.roles:
            personnel_manager = True

        role_model, create = CharacterRoles.objects.update_or_create(
            character=audit_char,
            defaults={
                "director": director,
                "accountant": accountant,
                "station_manager": station_manager,
                "personnel_manager": personnel_manager
            }
        )
        logger.debug(
            f"CT_TIME: {time.perf_counter() - _st} update_character_roles {character_id}"
        )

    except HTTPNotModified:
        logger.info(
            f"CT: No New roles for: {audit_char.character.character_name}"
        )

    audit_char.last_update_roles = timezone.now()
    audit_char.save()
    audit_char.is_active()

    return f"CT: Finished roles for: {audit_char.character.character_name}"


def update_character_mail_body(character_id, mail_message, force_refresh=False):
    CharacterAudit.objects.get(character__character_id=character_id)

    req_scopes = ['esi-mail.read_mail.v1']

    token = get_token(character_id, req_scopes)

    if not token:
        return False

    details = providers.esi.client.Mail.get_characters_character_id_mail_mail_id(
        character_id=character_id, mail_id=mail_message.mail_id,
        token=token.valid_access_token()
    ).result()

    mail_message.body = details.get('body')

    return mail_message


def update_character_mail_headers(character_id, force_refresh=False):
    # This function will deal with ALL mail related updates
    audit_char = CharacterAudit.objects.get(
        character__character_id=character_id)

    req_scopes = ['esi-mail.read_mail.v1']

    token = get_token(character_id, req_scopes)

    if not token:
        return False

    _current_eve_ids = list(
        EveName.objects.all().values_list('eve_id', flat=True))
    _current_mail_rec = list(
        MailRecipient.objects.all().values_list('recipient_id', flat=True))

    # Mail Labels
    labels = providers.esi.client.Mail.get_characters_character_id_mail_labels(character_id=character_id,
                                                                               token=token.valid_access_token()).result()

    for label in labels.get('labels'):
        MailLabel.objects.update_or_create(character=audit_char,
                                           label_id=label.get('label_id'),
                                           defaults={
                                               'label_name': label.get('name', None),
                                               'label_color': label.get('color', None),
                                               'unread_count': label.get('unread_count', None)
                                           })

    # Get all mail ids
    mail_ids = MailMessage.objects.filter(
        character=audit_char).values_list("mail_id", flat=True)
    last_id = None
    while True:
        if last_id is None:
            mail = providers.esi.client.Mail.get_characters_character_id_mail(character_id=character_id,
                                                                              token=token.valid_access_token()).result()
        else:
            mail = providers.esi.client.Mail.get_characters_character_id_mail(character_id=character_id,
                                                                              last_mail_id=last_id,
                                                                              token=token.valid_access_token()).result()
        if len(mail) == 0:
            # If there are 0 and this is not the first page, then we have reached the
            # end of retrievable mail.
            break

        messages = []
        m_l_map = {}
        m_r_map = {}
        failed_ids = set()
        stop = False
        for msg in mail:
            if msg.get('mail_id') == mail_ids:
                if not force_refresh:
                    break
                continue

            id_k = int(str(audit_char.character.character_id) +
                       str(msg.get('mail_id')))
            if msg.get('from', 0) not in _current_eve_ids:
                if msg.get('from', 0) not in failed_ids:
                    try:
                        EveName.objects.get_or_create_from_esi(msg.get('from'))
                        _current_eve_ids.append(msg.get('from', 0))
                    except Exception:
                        failed_ids.add(msg.get('from'))
                        pass
            msg_obj = MailMessage(character=audit_char, id_key=id_k, mail_id=msg.get('mail_id'), from_id=msg.get('from', None),
                                  is_read=msg.get('read', None), timestamp=msg.get('timestamp'),
                                  subject=msg.get('subject', None), body=msg.get('body', None))

            from_name_id = msg.get('from', None)
            if from_name_id in _current_eve_ids:
                msg_obj.from_name_id = msg.get('from', None)

            messages.append(msg_obj)

            if msg.get('labels'):
                labels = msg.get('labels')
                m_l_map[msg.get('mail_id')] = labels

            m_r_map[msg.get('mail_id')] = [(r.get('recipient_id'), r.get('recipient_type'))
                                           for r in msg.get('recipients')]
            last_id = msg.get('mail_id')

        msgs = MailMessage.objects.bulk_create(
            messages, batch_size=1000, ignore_conflicts=True)

        LabelThroughModel = MailMessage.labels.through
        lms = []
        for _msg in msgs:
            if _msg.mail_id in m_l_map:
                for label in m_l_map[_msg.mail_id]:
                    lm = LabelThroughModel(mailmessage_id=_msg.id_key,
                                           maillabel_id=MailLabel.objects.get(character=audit_char, label_id=label).pk)
                    lms.append(lm)

        LabelThroughModel.objects.bulk_create(lms, ignore_conflicts=True)

        RecipThroughModel = MailMessage.recipients.through
        rms = []
        for _msg in msgs:
            if _msg.mail_id in m_r_map:
                for recip, r_type in m_r_map[_msg.mail_id]:
                    recip_name = None
                    if r_type != "mailing_list":
                        if recip not in _current_eve_ids:
                            EveName.objects.get_or_create_from_esi(recip)
                            _current_eve_ids.append(recip)
                        recip_name = recip
                    if recip not in _current_mail_rec or force_refresh:
                        MailRecipient.objects.update_or_create(recipient_id=recip,
                                                               defaults={
                                                                   "recipient_name_id": recip_name,
                                                                   "recipient_type": r_type})
                        if not force_refresh:
                            _current_mail_rec.append(recip)

                    rm = RecipThroughModel(
                        mailmessage_id=_msg.id_key, mailrecipient_id=recip)
                    rms.append(rm)

        RecipThroughModel.objects.bulk_create(rms, ignore_conflicts=True)

        if stop is True:
            # Break the while loop if we reach the last mail message that is in the db.
            break

    audit_char.last_update_mails = timezone.now()
    audit_char.save()
    audit_char.is_active()

    if len(mail_ids) == 0:
        logger.debug(f"No new mails for {character_id}")
        return

    return mail_ids


def update_character_contacts(character_id, force_refresh=False):
    audit_char = CharacterAudit.objects.get(
        character__character_id=character_id)

    logger.debug(
        f"Updating Contacts for: {audit_char.character.character_name}"
    )

    req_scopes = ['esi-characters.read_contacts.v1']

    token = get_token(character_id, req_scopes)

    if not token:
        return False

    _current_eve_ids = list(
        EveName.objects.all().values_list('eve_id', flat=True)
    )

    try:
        labels = providers.esi_openapi.client.Contacts.GetCharactersCharacterIdContactsLabels(
            character_id=character_id,
            token=token
        ).result(
            force_refresh=force_refresh
        )

        _st = time.perf_counter()

        labels_to_create = []

        for label in labels:  # update labels
            _label_item = CharacterContactLabel.from_esi_model(
                audit_char, label)
            _label_item.build_id()
            labels_to_create.append(_label_item)

        CharacterContactLabel.objects.filter(character=audit_char).delete()
        CharacterContactLabel.objects.bulk_create(labels_to_create)

        logger.debug(
            f"CT_TIME: {time.perf_counter() - _st} CharacterContactLabel {character_id}"
        )
    except HTTPNotModified:
        logger.info(
            f"CT: No New old orders data for: {audit_char.character.character_name}"
        )

    try:
        contacts = providers.esi_openapi.client.Contacts.GetCharactersCharacterIdContacts(
            character_id=character_id,
            token=token
        ).result(
            force_refresh=force_refresh
        )

        ContactLabelThrough = CharacterContact.labels.through
        _contacts_to_create = []
        _through_to_create = []
        for contact in contacts:  # update contacts
            if contact.contact_id not in _current_eve_ids:
                EveName.objects.get_or_create_from_esi(
                    contact.contact_id
                )
                _current_eve_ids.append(contact.contact_id)

            _contact_item = CharacterContact.from_esi_model(
                audit_char, contact)

            _contacts_to_create.append(_contact_item)

            if contact.label_ids:  # add labels
                for _label in contact.label_ids:
                    _label_id = int(str(audit_char.id) + str(_label))
                    _lbl = ContactLabelThrough(
                        charactercontact_id=_contact_item.id,
                        charactercontactlabel_id=_label_id
                    )
                    _through_to_create.append(_lbl)

        CharacterContact.objects.filter(character=audit_char).delete()

        CharacterContact.objects.bulk_create(
            _contacts_to_create, ignore_conflicts=True)
        ContactLabelThrough.objects.bulk_create(
            _through_to_create, ignore_conflicts=True)

        logger.debug(
            f"CT_TIME: {time.perf_counter() - _st} update_character_contacts {character_id}"
        )

    except HTTPNotModified:
        logger.info(
            f"CT: No New contacts for: {audit_char.character.character_name}"
        )
        pass

    audit_char.last_update_contacts = timezone.now()
    audit_char.save()
    audit_char.is_active()

    return "CT: Completed contacts for: %s" % str(character_id)


def update_character_titles(character_id, force_refresh=False):
    audit_char = CharacterAudit.objects.get(
        character__character_id=character_id)

    logger.debug(
        f"Updating Title for: {audit_char.character.character_name}"
    )

    req_scopes = ['esi-characters.read_titles.v1']

    token = get_token(character_id, req_scopes)

    if not token:
        return False

    try:
        titles = providers.esi_openapi.client.Character.GetCharactersCharacterIdTitles(
            character_id=character_id,
            token=token
        ).result(
            force_refresh=force_refresh
        )

        _st = time.perf_counter()

        title_models = []
        for t in titles:  # update titles
            _title_item, _ = CharacterTitle.objects.update_or_create(
                corporation_id=audit_char.character.corporation_id,
                corporation_name=audit_char.character.corporation_name,
                title_id=t.title_id,
                defaults={
                    "title": strip_tags(t.name)
                }
            )

            title_models.append(_title_item.pk)
            audit_char.characterroles.titles.add(_title_item)

        if len(title_models) > 0:
            rem_tits = audit_char.characterroles.titles.all().exclude(pk__in=title_models)
            audit_char.characterroles.titles.remove(*rem_tits)
        else:
            audit_char.characterroles.titles.clear()

        logger.debug(
            f"CT_TIME: {time.perf_counter() - _st} update_character_titles {character_id}"
        )

    except HTTPNotModified:
        logger.info(
            f"CT: No New titles for: {audit_char.character.character_name}"
        )

    audit_char.last_update_titles = timezone.now()
    audit_char.save()
    audit_char.is_active()

    return "CT: Completed titles for: %s" % str(character_id)


def update_character_contracts(character_id, force_refresh=False):
    audit_char = CharacterAudit.objects.get(
        character__character_id=character_id)

    logger.debug(
        f"Updating Contracts for: {audit_char.character.character_name}"
    )

    req_scopes = ['esi-contracts.read_character_contracts.v1']

    token = get_token(character_id, req_scopes)

    new_contract_ids = []

    if not token:
        return False, []
    try:

        contracts = providers.esi_openapi.client.Contracts.GetCharactersCharacterIdContracts(
            character_id=character_id,
            token=token
        ).results(
            force_refresh=force_refresh
        )

        _st = time.perf_counter()

        contract_models_new = []
        contract_models_old = []

        eve_names = set()

        contract_ids = list(
            Contract.objects.filter(
                character=audit_char
            ).values_list(
                "contract_id",
                flat=True
            )
        )

        for c in contracts:  # update labels
            _contract_item = Contract.from_esi_model(audit_char, c)

            eve_names.add(c.assignee_id)
            eve_names.add(c.issuer_corporation_id)
            eve_names.add(c.issuer_id)
            eve_names.add(c.acceptor_id)

            if c.contract_id in contract_ids:
                contract_models_old.append(_contract_item)
            else:
                contract_models_new.append(_contract_item)

        EveName.objects.create_bulk_from_esi(list(eve_names))

        if len(contract_models_new) > 0:
            Contract.objects.bulk_create(
                contract_models_new,
                ignore_conflicts=True
            )

        if len(contract_models_old) > 0:
            Contract.objects.bulk_update(
                contract_models_old,
                fields=[
                    'date_accepted',
                    'date_completed',
                    'acceptor_id',
                    'date_expired',
                    'status',
                    'assignee_name',
                    'acceptor_name',
                    'issuer_corporation_name',
                    'issuer_name'
                ]
            )

        logger.debug(
            f"CT_TIME: {time.perf_counter() - _st} update_character_contracts {character_id}"
        )

        new_contract_ids = [c.contract_id for c in contract_models_new]
        if force_refresh:
            new_contract_ids += [c.contract_id for c in contract_models_old]

    except HTTPNotModified:
        logger.info(
            f"CT: No New Contracts for: {audit_char.character.character_name}"
        )

    audit_char.last_update_contracts = timezone.now()
    audit_char.save()
    audit_char.is_active()

    return "CT: Completed Contracts for: %s" % str(character_id), new_contract_ids


def update_character_contract_items(character_id, contract_id, force_refresh=False):
    audit_char = CharacterAudit.objects.get(
        character__character_id=character_id)

    logger.debug(
        f"Updating Contract Items for: {audit_char.character.character_name} - {contract_id}"
    )

    contract = Contract.objects.get(
        character=audit_char,
        contract_id=contract_id
    )

    req_scopes = ['esi-contracts.read_character_contracts.v1']

    token = get_token(character_id, req_scopes)

    if not token:
        return False
    try:
        contracts = providers.esi_openapi.client.Contracts.GetCharactersCharacterIdContractsContractIdItems(
            character_id=character_id,
            contract_id=contract_id,
            token=token
        ).results(
            force_refresh=force_refresh
        )

        _st = time.perf_counter()

        contract_models_new = []
        _types = set()

        for c in contracts:  # update labels
            _contract_item = ContractItem.from_esi_model(contract, c)
            _types.add(c.type_id)
            contract_models_new.append(_contract_item)

        EveItemType.objects.create_bulk_from_esi(list(_types))

        ContractItem.objects.bulk_create(
            contract_models_new,
            batch_size=1000,
            ignore_conflicts=True
        )

        logger.debug(
            f"CT_TIME: {time.perf_counter() - _st} update_character_contract_items {character_id}"
        )
    except HTTPNotModified:
        logger.info(
            f"CT: No New Contract items for: {audit_char.character.character_name}"
        )

    return f"CT: Completed Contract items {str(contract_id)} for: {str(character_id)}"



from random import random

from celery import chain as Chain, shared_task

from allianceauth.services.hooks import get_extension_logger
from allianceauth.services.tasks import QueueOnce

from ..models import CorporationAudit
from ..task_helpers import corp_helpers
from . import app_settings
from .utils import esi_error_retry, no_fail_chain

TZ_STRING = "%Y-%m-%dT%H:%M:%SZ"


logger = get_extension_logger(__name__)


@shared_task(
    bind=True,
    base=QueueOnce,
    name="corptools.tasks.update_corp_wallet"
)
@no_fail_chain
@esi_error_retry
def update_corp_wallet(self, corp_id, full_update=False, chain=[]):
    return corp_helpers.update_corp_wallet_division(corp_id, full_update=full_update)


@shared_task(
    bind=True,
    base=QueueOnce,
    name="corptools.tasks.update_corp_structures"
)
@no_fail_chain
@esi_error_retry
def update_corp_structures(self, corp_id, force_refresh=False, chain=[]):
    return corp_helpers.update_corp_structures(corp_id, force_refresh=force_refresh)


@shared_task(
    bind=True,
    base=QueueOnce,
    name="corptools.tasks.update_corp_assets"
)
@esi_error_retry
def update_corp_assets(self, corp_id, force_refresh=False, chain=[]):
    return corp_helpers.update_corp_assets(corp_id, force_refresh=force_refresh)


@shared_task(
    bind=True,
    base=QueueOnce,
    name="corptools.tasks.update_corp_pocos"
)
@no_fail_chain
@esi_error_retry
def update_corp_pocos(self, corp_id, chain=[]):
    return corp_helpers.update_corporation_pocos(corp_id)


@shared_task(
    bind=True,
    base=QueueOnce,
    name="corptools.tasks.update_corp_logins"
)
@no_fail_chain
@esi_error_retry
def update_corp_logins(self, corp_id, chain=[]):
    return corp_helpers.update_character_logins_from_corp(corp_id)


@shared_task(
    bind=True,
    base=QueueOnce,
    name="corptools.tasks.update_corp_starbases"
)
@no_fail_chain
@esi_error_retry
def update_corp_starbases(self, corp_id, force_refresh=False, chain=[]):
    return corp_helpers.fetch_starbases(corp_id, force_refresh=force_refresh)


@shared_task(
    bind=True,
    base=QueueOnce,
    name="corptools.tasks.update_corp_contracts"
)
@no_fail_chain
@esi_error_retry
def update_corp_contracts(self, corp_id, force_refresh=False, chain=[]):
    _, ids = corp_helpers.update_corporate_contracts(
        corp_id, force_refresh=force_refresh)

    _chain = []
    for id in ids:
        _chain.append(corp_helpers.update_corporate_contract_items.si(
            corp_id, id)
        )
    Chain(_chain).apply_async(priority=8)

    return "Completed Que of contract items for: %s" % str(corp_id)


@shared_task(
    bind=True,
    base=QueueOnce,
    name="corptools.tasks.update_corp_industry_jobs"
)
@no_fail_chain
@esi_error_retry
def update_corp_industry_jobs(self, corp_id: int, force_refresh: bool = False, chain=[]) -> str:
    return corp_helpers.update_corporation_industry_jobs(corp_id, force_refresh=force_refresh)


@shared_task(
    name="corptools.tasks.update_corp"
)
def update_corp(corp_id, force_refresh=False):
    corp = CorporationAudit.objects.get(corporation__corporation_id=corp_id)
    logger.info("Starting Updates for {}".format(
        corp.corporation.corporation_name))
    que = []
    que.append(update_corp_wallet.si(corp_id))
    que.append(update_corp_structures.si(corp_id, force_refresh=force_refresh))
    que.append(update_corp_starbases.si(corp_id, force_refresh=force_refresh))
    que.append(update_corp_assets.si(corp_id, force_refresh=force_refresh))
    que.append(update_corp_pocos.si(corp_id))
    que.append(update_corp_contracts.si(corp_id))
    que.append(update_corp_logins.si(corp_id))
    que.append(update_corp_industry_jobs.si(
        corp_id, force_refresh=force_refresh))
    Chain(que).apply_async(priority=6)


@shared_task(
    name="corptools.tasks.update_all_corps"
)
def update_all_corps(force_refresh=False):
    corps = CorporationAudit.objects.all().select_related('corporation')
    for corp in corps:
        countdown = 1 if force_refresh else random()*app_settings.CT_TASK_SPREAD_DELAY*2
        update_corp.apply_async(
            args=[corp.corporation.corporation_id],
            kwargs={"force_refresh": force_refresh},
            countdown=countdown
        )



# Generated by Django 4.2.16 on 2025-06-07 00:33

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('corptools', '0121_charassetcoordiante'),
    ]

    operations = [
        migrations.AddField(
            model_name='characteraudit',
            name='total_logins',
            field=models.IntegerField(blank=True, default=None, null=True),
        ),
    ]




from django.db import models

from corptools.task_helpers import sanitize_location_flag

from .audits import (
    CharacterAudit, CorporationAudit, CorptoolsConfiguration, EveLocation,
)
from .eve_models import EveItemType


class Asset(models.Model):
    id = models.BigAutoField(primary_key=True)
    blueprint_copy = models.BooleanField(null=True, default=None)
    singleton = models.BooleanField()
    item_id = models.BigIntegerField()
    location_flag = models.CharField(max_length=50)
    location_id = models.BigIntegerField()
    location_type = models.CharField(max_length=25)
    quantity = models.IntegerField()
    type_id = models.IntegerField()
    type_name = models.ForeignKey(
        EveItemType, on_delete=models.SET_NULL, null=True, default=None)
    location_name = models.ForeignKey(
        EveLocation, on_delete=models.SET_NULL, null=True, default=None)

    # extra's
    name = models.CharField(max_length=255, null=True, default=None)

    class Meta:
        abstract = True
        indexes = [
            models.Index(fields=['location_id']),
            models.Index(fields=['item_id']),
        ]


class CorpAsset(Asset):
    corporation = models.ForeignKey(CorporationAudit, on_delete=models.CASCADE)

    def __str__(self):
        return '{2} {0}x{1} ({3} / {4})'.format(self.type_id, self.quantity, self.corporation,
                                                self.location_id, self.location_type)

    @classmethod
    def get_visible(cls, user):
        corps_vis = CorporationAudit.objects.visible_to(user)
        if user.has_perm("corptools.holding_corp_assets"):
            corps_holding = CorptoolsConfiguration.get_solo().holding_corp_qs()
            corps_vis = corps_vis | corps_holding

        return cls.objects.filter(corporation__in=corps_vis)

    @classmethod
    def from_esi_model(cls, corporation: CorporationAudit, model):
        return cls(
            corporation=corporation,
            blueprint_copy=model.is_blueprint_copy,
            singleton=model.is_singleton,
            item_id=model.item_id,
            location_flag=sanitize_location_flag(
                model.location_flag
            ),
            location_id=model.location_id,
            location_type=model.location_type,
            quantity=model.quantity,
            type_id=model.type_id,
            type_name_id=model.type_id
        )


class CharacterAsset(Asset):
    character = models.ForeignKey(CharacterAudit, on_delete=models.CASCADE)

    def __str__(self):
        return '{2} {0}x{1} ({3} / {4})'.format(
            self.type_id,
            self.quantity,
            self.character,
            self.location_id,
            self.location_type
        )

    @classmethod
    def from_esi_model(cls, character: CharacterAudit, model):
        return cls(
            character=character,
            blueprint_copy=model.is_blueprint_copy,
            singleton=model.is_singleton,
            item_id=model.item_id,
            location_flag=sanitize_location_flag(
                model.location_flag
            ),
            location_id=model.location_id,
            location_type=model.location_type,
            quantity=model.quantity,
            type_id=model.type_id,
            type_name_id=model.type_id
        )

    @classmethod
    def from_esi_location(cls, character: CharacterAudit, ship, location):
        location_id = location.solar_system_id
        location_flag = "solar_system"
        location_type = "unlocked"
        if location.structure_id is not None:
            location_id = location.structure_id
            location_flag = "item"
            location_type = "hangar"
        elif location.station_id is not None:
            location_id = location.station_id
            location_flag = "station"
            location_type = "hangar"

        return cls(
            character=character,
            singleton=True,
            item_id=ship.ship_item_id,
            location_flag=sanitize_location_flag(
                location_flag
            ),
            location_id=location_id,
            location_type=location_type,
            quantity=1,
            type_id=ship.ship_type_id,
            type_name_id=ship.ship_type_id
        )


class AssetCoordiante(models.Model):
    item = models.OneToOneField(
        CorpAsset,
        on_delete=models.CASCADE,
        primary_key=True,
        related_name="coordinate"
    )
    x = models.FloatField()
    y = models.FloatField()
    z = models.FloatField()


class CharAssetCoordiante(models.Model):
    item = models.OneToOneField(
        CharacterAsset,
        on_delete=models.CASCADE,
        primary_key=True,
        related_name="coordinate"
    )
    x = models.FloatField()
    y = models.FloatField()
    z = models.FloatField()



# Generated by Django 4.2.16 on 2025-06-13 09:15

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('corptools', '0123_characterwalletjournalentry_processed_and_more'),
    ]

    operations = [
        migrations.CreateModel(
            name='CharacterBountyEvent',
            fields=[
                ('id', models.AutoField(auto_created=True,
                 primary_key=True, serialize=False, verbose_name='ID')),
                ('message', models.TextField()),
                ('entry', models.OneToOneField(on_delete=django.db.models.deletion.CASCADE,
                 related_name='msg', to='corptools.characterwalletjournalentry')),
            ],
        ),
        migrations.CreateModel(
            name='CharacterBountyStat',
            fields=[
                ('id', models.AutoField(auto_created=True,
                 primary_key=True, serialize=False, verbose_name='ID')),
                ('qty', models.IntegerField()),
                ('event', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE,
                 related_name='stats', to='corptools.characterbountyevent')),
                ('type_name', models.ForeignKey(
                    on_delete=django.db.models.deletion.CASCADE, to='corptools.eveitemtype')),
            ],
        ),
    ]



import logging

from ninja import NinjaAPI

from django.core.cache import cache
from django.utils.translation import gettext as _

from allianceauth.eveonline.tasks import (
    update_character as eve_character_update,
)

from corptools import app_settings, models
from corptools.api import schema
from corptools.api.helpers import get_alts_queryset, get_main_character
from corptools.tasks import character

logger = logging.getLogger(__name__)


class RefreshApiEndpoints:

    tags = ["Characters"]

    def __init__(self, api: NinjaAPI):

        @api.post(
            "characters/refresh",
            response={200: schema.Message, 304: str, 403: str, 404: str},
            tags=self.tags
        )
        def post_characters_refresh(request, character_id: int):
            if character_id == 0:
                character_id = request.user.profile.main_character.character_id
            response, main = get_main_character(request, character_id)

            if not response:
                return 403, _("Permission Denied")

            # and not request.user.is_superuser:
            if cache.get(f"refresh-block-{character_id}", False):
                logger.error(
                    f"GO AWAY! Already Requested! {character_id} {request.user.username}")
                return 200, {"message": "GO AWAY! Already Requested!"}

            characters = get_alts_queryset(main)
            if character_id in characters.values_list('character_id', flat=True):
                character.update_char_assets.apply_async(
                    args=[character_id],
                    kwargs={
                        "force_refresh": True
                    },
                    priority=4)
                cache.set(f"refresh-block-{character_id}", 1, 60*5)
                return 200, {"message": "Requested Update!"}
            return 404, "Not Found!"

        @api.post(
            "account/refresh",
            response={200: schema.Message, 403: str},
            tags=self.tags
        )
        def post_account_refresh(request, character_id: int):
            if character_id == 0:
                character_id = request.user.profile.main_character.character_id
            response, main = get_main_character(request, character_id)

            if not response:
                return 403, _("Permission Denied")

            characters = get_alts_queryset(main)

            force = app_settings.CT_USERS_CAN_FORCE_REFRESH or request.user.is_superuser

            # and not request.user.is_superuser:
            if cache.get(f"refresh-block-account-{character_id}", False):
                logger.error(
                    f"GO AWAY! Already Requested! {character_id} {request.user.username}")
                return 200, {"message": "GO AWAY! Already Requested!"}

            for cid in set(list(characters.values_list('character_id', flat=True))):
                character.update_character.apply_async(
                    args=[cid],
                    kwargs={
                        "force_refresh": force
                    },
                    priority=4)
                cache.set(f"refresh-block-account-{character_id}", 1, 60*30)
                eve_character_update.apply_async(
                    args=[cid], priority=4)
            return 200, {"message": "Requested Updates!"}



# Generated by Django 4.2.16 on 2025-06-09 10:31

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('corptools', '0122_characteraudit_total_logins'),
    ]

    operations = [
        migrations.AddField(
            model_name='characterwalletjournalentry',
            name='processed',
            field=models.BooleanField(default=False),
        ),
        migrations.AddField(
            model_name='corporationwalletjournalentry',
            name='processed',
            field=models.BooleanField(default=False),
        ),
    ]
